import numpy as np
import pickle

from ..lib import bspline
from . import etc, reindex
from ..dataset import TrainingDataset


try:
    from sksparse.cholmod import cholesky_AAt
except ImportError:
    from scikits.sparse.cholmod import cholesky_AAt

NORM_SNSIM = 290090.4311080577


class Simulator(object):
    r"""
    From data generator model flux and error
    on measurements.

    Flux are created with SALT2.4
    Flux error come from different sources.
    For ideal simulation there are simulated, always constant  fraction of flux of spectral data.
    For LSST, flux error are calculated from an ETC, otherwise they are constant fraction of flux.

    .. math::
        \sigma^2 = \left\{
         \sigma_{Err}^2 + \sigma_{mod}^2 + \sigma_{seuil}^2 \quad for spectral data \\
        \sigma_{Err}^2 + \sigma_{mod}^2 + \sigma^2_{\eta} + \sigma_{\kappa}^2 + \sigma_{seuil}^2  \quad otherwise


    Return lc and sp array, with the conventional format.

    Attributes
    ----------
    model_func : nacl.models.salt.SALT2Like
        Model to generated data.
    degree : int
        Degree of the spectral recalibration polynomial function.
    normalization_band_name : str
        Band name for normalization of the model.
    init_from_salt2_file : str
        File salt2, created by script 'nacl.simulation.make_salt2_npz.py'.
    disconnect_sp :
        Whether spectral data are considered.

    variance_model_func : nacl.models.variancemodels
        Error snake model to add as a noise in the simulated data.
    gamma_init : numpy.array ou float
        Parameters used to initialize the error snake.
    seed : int
        Seed for cadence.

    CS : nacl.models.ColorScatter
        Model of the color scatter to add as a noise.

    sigma_kappa_init : numpy.array
        Color scatter parameters to add in simulation.
    eta_covmatrix : float or array
        Covariance of calibration uncertainty to add in simulation as a noise.

    sig_calib : float or numpy.array
        Dispersion of the :math:`\eta` parameters.

    kappa_init : numpy.array
        Color scatter parameter used in the simulation.

    color_scatter_eval : numpy.array
        Evaluation of color scatter for each light curve.
    kappa : numpy.array
        :math:`\kappa` drawn according to normal distribution of standard deviation equal to poly.

    sig : numpy.array
        Variance model evaluation (quadratic sum of measurement error and error snake).

    calib : numpy.array
        Value of :math:`\eta` of each light curve data point.
    eta_band : dict
        Value of :math:`\eta` parameters used model evaluation of the simulation distributed by centred
        gaussian of standard deviation equal to  sig_calib for each band.

    lambda_c : numpy.array
        Mean wavelength of the filter in the SN-restframe.

    lambda_c_red : numpy.array
        Reduced mean wavelength of the filter in the SN-restframe.

    data_generator : nacl.jla.generator or nacl.ideal.generator or nacl.k21.generator
        Cadence and SNe information.

    sigma_lc : numpy.array
        Dispersion of photometric data
    filterpath : str
        Path to the filters files repository in nacl.dev/data/
    Sne ; numpy.array
        SNe names
    lc : numpy.rec.array
        Light curves data with data_type as type.
    sp : numpy.rec.array
        Spectral data, with data_type as type.

    idx_sp : numpy.array
        Spectral data index cuts.
    idx_lc : numpy.array
        Photometric data index cuts.

    trainingDataset
    """
    def __init__(self, data_generator, model_func, variance_model_func, 
                 init_from_salt2_file, degree=3, normalization_band_name='SWOPE::B',
                 noise=True, seed=0, gamma_init=0.0001, sigma_kappa_init=None,
                 disconnect_sp=False, eta_covmatrix=None, color_scatter_func=None,
                 init_filename=None):
        r"""
        Constructor.

        Parameters
        ----------
        data_generator : nacl.jla.generator or nacl.ideal.generator or nacl.k21.generator
            Cadence and SNe information.
        model_func : nacl.models.salt.SALT2Like
            Model to generated data.
        variance_model_func : nacl.models.variancemodels
            Error snake model to add as a noise in the simulated data.
        degree : int
            Degree of the spectral recalibration polynomial function.
        normalization_band_name : str
            Band name for normalization of the model.
        init_from_salt2_file : str
            File salt2, created by script 'nacl.simulation.make_salt2_npz.py'.
        disconnect_sp :
            Whether spectral data are considered.
        noise : bool
            Whether error snake is considered in simulation.
        gamma_init : numpy.array ou float
            Parameters used to initialize the error snake.
        seed : int
            Seed for cadence.
        color_scatter_func : nacl.models.ColorScatter
            Model of the color scatter to add as a noise.
        sigma_kappa_init : numpy.array
            Color scatter parameters to add in simulation.
        eta_covmatrix : float or array
            Covariance of calibration uncertainty to add in simulation as a noise.
        init_filename : str
            If simulation is already stored somewhere.
        """
        self.model_func = model_func
        self.degree = degree
        self.seed = seed
        np.random.seed(self.seed)
        self.gamma_init = gamma_init

        self.sigma_kappa_init = sigma_kappa_init
        
        self.eta_covmatrix = eta_covmatrix
        self.init_from_salt2_file = init_from_salt2_file
        self.variance_model_func = variance_model_func
        self.normalization_band_name = normalization_band_name
        self.disconnect_sp = disconnect_sp

        self.sig_calib = None
        self.kappa = None
        
        if init_filename is not None:
            self.init_from_data(init_filename)
        else:
            self.sig, self.seed = None, None
            self.calib, self.eta_band, self.lambda_c_red, self.lambda_c, self.color_scatter_eval = None, None, None, None, None
            self.kappa_init, self.sigma_lc = None, None
            self.data_generator = data_generator
            try:
                self.data_generator.lsst_simu
            except AttributeError:
                self.data_generator.lsst_simu = False

            try:
                self.Sne = data_generator.Sne[np.unique(self.lc['sn_id'])]
            except AttributeError:
                self.Sne = None
            self.filterpath = self.data_generator.trainingDataset.filterpath
            self.model = self.model_func(self.data_generator.trainingDataset,
                                         init_from_salt2_file=None, spectrum_recal_degree=self.degree,
                                         normalization_band_name=self.normalization_band_name)
            
            self.sigma_kappa_init = sigma_kappa_init
            self.CS = None
            if (self.sigma_kappa_init is not None) & (color_scatter_func is not None):
                self.CS = color_scatter_func(self.model.lambda_c_red, self.sigma_kappa_init)
                
            self.model_pars_initialisation(self.init_from_salt2_file)

            self.lc = self.data_generator.trainingDataset.lc_data
            self.sp = self.data_generator.trainingDataset.spec_data

            self.idx_sp, self.idx_lc = self.data_cuts()
            self.lc, self.sp = self.lc[self.idx_lc], self.sp[self.idx_sp]
            
            self.snInfo = self.data_generator.snInfo[np.unique(self.lc['sn_id'])]
            
            self.recreate_model_from_new_data_set()

            if np.unique(self.lc['sn_id']).shape[0] != self.lc['sn_id'][-1]+1:
                raise ValueError('All lcs from a SN have been remove')

            self.make_data()
            print(f"LC: {self.lc['i'][-1]+1} should be {len(np.unique(self.lc['i']))}")
            print(f"SP: {self.sp['i'][-1]+1} should be {len(np.unique(self.sp['i']))}")
            self.recreate_model_from_new_data_set()

            print(f"LC: {self.lc['i'][-1]+1} should be {len(np.unique(self.lc['i']))}")
            print(f"SP: {self.sp['i'][-1]+1} should be {len(np.unique(self.lc['i']))+len(np.unique(self.sp['i']))}")

            if noise:
                self.variance_model = self.variance_model_func(self.model, gamma_init=self.gamma_init,
                                                               # kappa_init = self.kappa_init,
                                                               disconnect_sp=self.disconnect_sp)
                
                self.adding_noise()

        self.trainingDataset = TrainingDataset(self.lc, self.sp, self.snInfo,
                                               filterpath=self.filterpath)

    def adding_noise(self):
        """
        Add error snake to as a noise to the flux simulated.
        """
        var = self.variance_model(self.variance_model.pars.free, self.model.pars.free)
        sigma = np.sqrt(var)
        print("Adding noise \n")
        self.sig = np.random.normal(0, sigma)
        self.lc['Flux'] += self.sig[self.lc['i']]
        self.sp['Flux'] += self.sig[self.sp['i']]
        
    def recreate_model_from_new_data_set(self):
        """
        Recreate a model and initiated its parameters from new data set, where cuts were apply,
        and index (of light_curve, spectra, sn) have changed.
        """
        if self.kappa is not None:
            idx = np.unique(self.lc['lc_id']).copy()
            self.kappa = self.kappa[idx]
        
        self.lc['lc_id'] = reindex.reindex(self.lc['lc_id'])
        self.lc['band_id'] = reindex.reindex(self.lc['band_id'])
        self.sp['spec_id'] = reindex.reindex(self.sp['spec_id'])

        n_lc = len(self.lc['i'])
        n_sp = len(self.sp['i'])
        self.lc['i'] = np.arange(n_lc)
        self.sp['i'] = n_lc + np.arange(n_sp)

        ts = TrainingDataset(self.lc, self.sp, self.snInfo)
        self.model = self.model_func(ts, init_from_salt2_file=None,
                                     spectrum_recal_degree=self.degree,
                                     normalization_band_name=self.normalization_band_name)
        self.model_pars_initialisation(self.init_from_salt2_file)

    def model_pars_initialisation(self, filename):
        r"""
        Initiated model parameters, global parameters are salt2's.
        Sne parameters are taken from data_generator.snInfo.
        Photometric calibration parameters are zero and spectroscopic are 1 !!!

        Parameters
        ----------
        filename : str
            File salt2, created by script 'nacl.simulation.make_salt2_npz.py'.
        """
        self.model.pars['X0'].full[:] = self.data_generator.snInfo['x0']
        print(f"model init X1 {self.data_generator.snInfo['x1'].mean()}; {self.data_generator.snInfo['x1'].std()}")
        self.model.pars['X1'].full[:] = self.data_generator.snInfo['x1']
        self.model.pars['c'].full[:] = self.data_generator.snInfo['c']
        self.model.pars['tmax'].full[:] = self.data_generator.snInfo['tmax']

        if self.eta_covmatrix is not None:
            if self.sig_calib is None:
                if type(self.eta_covmatrix) == float:
                    self.sig_calib = np.sqrt(self.eta_covmatrix) * np.ones(len(self.model.bands))
                else:
                    self.sig_calib = np.sqrt(self.eta_covmatrix)
                eta = np.random.normal(0, self.sig_calib)
                self.eta_band = {self.model.bands[i]: eta_i for i, eta_i in enumerate(eta)}

            self.calib = [self.eta_band[bd] for bd in self.model.bands]
                
        else:
            self.calib = 0.
        self.model.pars['eta_calib'].full[:] = self.calib 

        self.lambda_c = self.model.lambda_c 
        self.lambda_c_red = self.model.lambda_c_red
        pcl = self.model.pars.full[self.model.pars['CL'].indexof()]
        cl_i, _ = self.model.color_law(self.lambda_c, pcl)
        
        if (self.kappa is None) & (self.sigma_kappa_init is not None):
            self.color_scatter_eval = self.CS(self.CS.pars.free)
            self.kappa = np.random.normal(0, self.color_scatter_eval)
            self.model.pars['kappa_color'].full[:] = self.kappa

        elif self.sigma_kappa_init is None:
            self.model.pars['kappa_color'].full[:] = 0.
            
        else:
            self.model.pars['kappa_color'].full[:] = self.kappa

        self.kappa_init = self.model.pars['kappa_color'].full[:].copy()
        f = np.load(filename)
        phase_grid = f['phase_grid']
        wl_grid = f['wl_grid']
        basis = bspline.BSpline2D(wl_grid, phase_grid, x_order=4, y_order=4)
        
        xx = np.linspace(wl_grid[0], wl_grid[-1], basis.bx.nj)
        yy = np.linspace(phase_grid[0], phase_grid[-1], basis.by.nj)
        
        x, y = np.meshgrid(xx, yy)
        x, y = x.ravel(), y.ravel()
        basis_eval = self.model.basis.eval(x, y)
        factor = cholesky_AAt(basis_eval.T, beta=1e-10)

        surface_0, surface_1, color_law = factor(basis_eval.T * f['M0']), factor(basis_eval.T * f['M1']), f['CL_pars']
        self.model.pars['M0'][:] = surface_0
        self.model.pars['M1'][:] = surface_1
        self.model.pars['CL'][:] = color_law
        self.model.norm = self.model.normalization(band_name=self.normalization_band_name)

    def make_photometric_sigma(self, sigmalc):
        r"""
        Make the photometric error measurement if the surveys is a simulation of LSST,
        it will use the error simulation of saunerie psf
        Else it used sigmalc, from Data generator.

        Parameters
        ----------
        sigmalc : float
            Part of the photometric flux to be the error measurements.
        """
        if self.data_generator.lsst_simu:
            zlc = self.lc['ZHelio']
            n_lc = len(zlc)
            exptime = np.zeros(n_lc)
            flux_sky = np.zeros(n_lc)
            seeing = np.zeros(n_lc)

            exptime[np.where(zlc > 0.4)[0]] = 600
            exptime[np.where(zlc <= 0.4)[0]] = 30

            lsst = etc.find('LSST')
            lsst.dump()
            for ib in range(len(lsst.data['band'])):
                idx = np.where(self.lc['Filter'] == lsst.data['band'][ib].encode('UTF-8'))[0]
                flux_sky[idx] = lsst.data['flux_sky'][ib]
                seeing[idx] = lsst.data['iq'][ib]

            band = np.array([i.decode('UTF-8') for i in self.lc['Filter']])
            
            variance = lsst.flux_variance(exptime=exptime, flux=self.lc['Flux']*NORM_SNSIM,
                                          skyflux=flux_sky, seeing=seeing,
                                          band=band)
            self.sigma_lc = np.sqrt(variance)/np.abs(self.lc['Flux']*NORM_SNSIM)
        else:
            if type(sigmalc) == np.ndarray:
                sigmalc = sigmalc[self.idx_lc]
            self.sigma_lc = sigmalc

    def make_data(self):
        r"""
        Create flux from the model evaluation.
        Measurement error are the product of the flux and
        the sigma array defined in the data_generator.
        """
        res = self.model(self.model.pars.free, jac=False)
        self.sp['Flux'] = res[len(self.lc):] 

        if type(self.data_generator.sigma_sp) == float:
            sig = self.data_generator.sigma_sp
        elif type(self.data_generator.sigma_sp) == np.ndarray:
            sig = self.data_generator.sigma_sp[self.idx_sp]
        
        self.sp['FluxErr'] = np.abs(sig * self.sp['Flux'])
        
        # Offset per SPs ... otherwise data point could be very low with sigma below...
        for isp in np.unique(self.sp['spec_id']):
            idx_sp = self.sp['spec_id'] == isp
            offset_sp = self.sp[idx_sp]['Flux'].max() * 5e-3  # 4
            self.sp['FluxErr'][idx_sp] = np.sqrt(self.sp['FluxErr'][idx_sp] ** 2 + offset_sp ** 2)

        if (np.isinf(self.sp['Flux']) ^ np.isinf(self.sp['FluxErr'])).sum() > 0:
            print('Inf in spectral flux set to zeros')
            idx_inf = np.where((np.isinf(self.sp['Flux']) ^ np.isinf(self.sp['FluxErr'])))[0]
            self.sp['Flux'][idx_inf] = 0.
            self.sp['FluxErr'][idx_inf] = 100.

        if (np.isnan(self.sp['Flux']) ^ np.isnan(self.sp['FluxErr'])).sum() > 0:
            print('NaN in spectral flux set to zeros')
            idx_nan = np.where((np.isnan(self.sp['Flux']) ^ np.isnan(self.sp['FluxErr'])))[0]
            self.sp['Flux'][idx_nan] = 0.
            self.sp['FluxErr'][idx_nan] = 100.
    
        self.sp = self.sp[self.sp['FluxErr'] != 100]
        self.lc['Flux'] = res[:len(self.lc)]  # lc_eval()
        
        self.make_photometric_sigma(self.data_generator.sigma_lc)
        self.lc['FluxErr'] = np.abs(self.sigma_lc) * np.abs(self.lc['Flux'])

        # Offset per LCs ... otherwise data point could be very low with sigma below...
        for ilc in np.unique(self.lc['lc_id']):
            idx_lc = self.lc['lc_id'] == ilc
            offset_lc = self.lc[idx_lc]['Flux'] * 5e-3
            self.lc['FluxErr'][idx_lc] = np.sqrt(self.lc['FluxErr'][idx_lc] ** 2 + offset_lc ** 2)
            
        if (np.isinf(self.lc['Flux']) ^ np.isinf(self.lc['FluxErr'])).sum() > 0:
            print('Inf in photo flux set to zeros')
            idx_inf = np.where((np.isinf(self.lc['Flux']) ^ np.isinf(self.lc['FluxErr'])))[0]
            self.lc['Flux'][idx_inf] = 0.
            self.lc['FluxErr'][idx_inf] = 100.

        if (np.isnan(self.sp['Flux']) ^ np.isnan(self.sp['FluxErr'])).sum() > 0:
            print('NaN in photo flux set to zeros')
            idx_nan = np.where((np.isnan(self.sp['Flux']) ^ np.isnan(self.sp['FluxErr'])))[0]
            self.lc['Flux'][idx_nan] = 0.
            self.lc['FluxErr'][idx_nan] = 100.
            
        self.lc = self.lc[self.lc['FluxErr'] != 100]

    def data_cuts(self):
        r"""
        Remove data out of the model definition. 
        If a date of observation is out of phase range, or wavelength out of wavelength range, 
        data point(s) is(are) removed.

        Returns
        -------
        idx_sp : numpy.array
            Spectral data index cuts.
        idx_lc : numpy.array
            Photometric data index cuts.
        """
        blue_cutoff, red_cutoff = self.model.wl_range
        early_cutoff, late_cutoff = self.model.phase_range

        # for JLA and K21 need cutoff to be less than model def;
        # in order to degrade tmax and fit
        early_cutoff += 5
        late_cutoff -= 5
        
        lc = self.lc
        sp = self.sp

        # lc
        bands = np.unique(lc['Filter'])
        lambda_range = np.linspace(1500, 11000, int(1e5))
        min_lambda = np.zeros(len(lc['Filter']))
        max_lambda = np.zeros(len(lc['Filter']))
        for bd in bands:
            idx = lc['Filter'] == bd
            ff = self.model.filter_db.transmission_db[bd.decode('UTF-8')].func(lambda_range)
            idxx = np.where(ff > ff.max()*0.01)
            mmin, mmax = lambda_range[idxx[0][0]], lambda_range[idxx[0][-1]]
            min_lambda[idx] = mmin/(1+lc['ZHelio'][idx])
            max_lambda[idx] = mmax/(1+lc['ZHelio'][idx])
        date_lc = (lc['Date'] - self.data_generator.snInfo['tmax'][lc['sn_id']]) / (1. + lc['ZHelio'])
        idx_ph_lc = (date_lc > early_cutoff) & (date_lc < late_cutoff)
        idx_wl_lc = (min_lambda > blue_cutoff) & (max_lambda < red_cutoff)
        idx_lc = idx_ph_lc & idx_wl_lc

        # sp
        wl_sp = sp['Wavelength'] / (1. + sp['ZHelio'])
        idx_wl_sp = (wl_sp > blue_cutoff) & (wl_sp < red_cutoff)
        date_sp = (sp['Date'] - self.data_generator.snInfo['tmax'][sp['sn_id']]) / (1. + sp['ZHelio'])
        idx_ph_sp = (date_sp > early_cutoff) & (date_sp < late_cutoff)
        
        idx_sp = idx_ph_sp & idx_wl_sp
        return idx_sp, idx_lc

    def init_from_data(self, filename):
        """
        Initiated light curve and spectra array from a stored file.

        Parameters
        ----------
        filename : str
            If simulation is already stored somewhere.
        """
        self.seed = 0
        dict_data = pickle.load(open(filename, "rb"))
        
        self.lc = dict_data['lc']
        self.sp = dict_data['spec']
        self.snInfo = dict_data['snInfo']
        self.filterpath = dict_data['filterpath']        
        if self.gamma_init != 0.:
            
            sig = dict_data['sig']
            self.lc['Flux'] -= sig[self.lc['i']]
            self.sp['Flux'] -= sig[self.sp['i']]
            flux = np.hstack((self.lc['Flux'], self.sp['Flux']))
            flux_err = np.hstack((self.lc['FluxErr'], self.sp['FluxErr']))
            
            val = flux_err ** 2 + (self.gamma_init * flux) ** 2
            if self.kappa_init != 0:
                print('kappa in generation')
                val += np.hstack((self.kappa_init * np.ones_like(self.lc['lc_id']),
                                  np.zeros_like(self.sp['i'])))**2
            sigma = np.sqrt(val)
            
            print("Adding New noise \n")
            if type(self.seed) == int:
                np.random.seed(self.seed)
            self.sig = np.random.normal(0, sigma)

            self.lc['Flux'] += self.sig[self.lc['i']]
            self.sp['Flux'] += self.sig[self.sp['i']]            



#!/usr/bin/env python
"""
Implements functions to describe and analyse atmospheric transmission.

In particular, re-implements the model of Buton et al. 2013, with code stolen from
CVS_SNf/Analysis/AtmosphericExtinction/Scripts/pyExtinction/AtmosphericExtinction.py


The anciliary data is to be found in $SALTPATH/atmospheric_model/

"""

import numpy
# temporary change, until pyfits is discontinued 
try:
    import astropy.io.fits as pyfits
except:
    import pyfits

import copy
    
from . import interpolation
from . import fitparameters
from ..util import saltpath

#- Utility I/O scripts to read atmospheric data parameters from SNf files

def read_snf_atmospheric_model(filename):
    """
    Reads the atmospheric transmission data from a file with FClass 625 and
    XFclass 610 (Non Photometric solution)
    of XFclass 600 (Photometric_transmission)
    """


    f_fits = pyfits.open(filename)
    header = f_fits[1].header

    atmospheric_transmission = atmosphericmodel.Buton_atmosphere(airmass=1.,
                                                                 pressure=header.get("PRESSURE"),
                                                                 ozone_column=header.get("OZ_INT"),
                                                                 aerosol_optical_depth=header.get("AE_TAU"),
                                                                 aerosol_exponent=header.get("AE_ANG"),
                                                                 ozone_template_path=None,
                                                                 keys=f_fits[1].header)

    return atmospheric_transmission


def read_snf_telluric_model(filename):
    """
    Reads the atmospheric telluric data and creates a telluric transmission model

    The file should be Fclass 620 XFclass 600

    """
    #- read the atmospheric transmission data

    f_fits = pyfits.open(filename)
    header = f_fits[1].header
    
    telluric_transmission = atmosphericmodel.Buton_telluric(airmass=1.,
                                                            H2Ow_power=header.get("TELLPOW0"),
                                                            H2Os_power=header.get("TELLPOW1"),
                                                            O2_power=header.get("TELLPOW2"),
                                                            H2Ow_intensity=header.get("TELLINT0"),
                                                            H2Os_intensity=header.get("TELLINT1"),
                                                            O2_intensity=header.get("TELLINT2"),                 
                                                            telluric_template_path=None,
                                                            keys=f_fits[1].header)        
    return telluric_transmission


def read_snf_full_model(atmospheric_filename, telluric_filename):
    """
    Reads a full atmospheric model from an atmospheric model file
    625_610 or 625_600
    and a telluric file
    620_600
    """
    f_fits_atmospheric = pyfits.open(atmospheric_filename)
    header_atmospheric = f_fits_atmospheric[1].header

    f_fits_telluric = pyfits.open(telluric_filename)
    header_telluric = f_fits_telluric[1].header


    keys = {}
    for item in list(f_fits_atmospheric[1].header.items()):
        keys[item[0]] = item[1]
    for item in list(f_fits_telluric[1].header.items()):
        keys[item[0]] = item[1]
    
    transmission_model = atmosphericmodel.Buton(airmass=1.,
                                                pressure=header_atmospheric.get("PRESSURE"),
                                                ozone_column=header_atmospheric.get("OZ_INT"),
                                                aerosol_optical_depth=header_atmospheric.get("AE_TAU"),
                                                aerosol_exponent=header_atmospheric.get("AE_ANG"),
                                                ozone_template_path=None,
                                                H2Ow_power=header_telluric.get("TELLPOW0"),
                                                H2Os_power=header_telluric.get("TELLPOW1"),
                                                O2_power=header_telluric.get("TELLPOW2"),
                                                H2Ow_intensity=header_telluric.get("TELLINT0"),
                                                H2Os_intensity=header_telluric.get("TELLINT1"),
                                                O2_intensity=header_telluric.get("TELLINT2"),                 
                                                telluric_template_path=None,
                                                keys=keys)        
    

    return transmission_model


#- Atmospheric model from Buton et al.
#- Directly usable with SNf fits keywords

def buton_base_rayleigh(wavelength):
    """Rayleigh extinction from Hansen & Travis (1974).

    Implementation stolen from CVS_SNf/Analysis/AtmosphericExtinction/Scripts/pyExtinction/AtmosphericExtinction.py

    wavelength: wavelength vector [AA]
    
    return: Rayleigh extinction [mag/airmass] without the pressure term
            pressure: effective surface pressure [mbar]

    """
    EXT2OPT = .92103403719761834  # LOG10/2.5 = Extinction to opt. thickness
    
    lm = wavelength * 1e-4                # Wavelength from A to microns
    
    # Optical depth
    tau = 0.008569 / lm ** 4 * (1 + 0.0113 / lm ** 2 + 0.00013 / lm ** 4)
    tau *= 1. / (1013.25 * EXT2OPT) 
    
    return tau


class Buton_atmosphere(object):
    def __init__(self, airmass=1.,
                 pressure=616.,
                 ozone_column=257.,
                 aerosol_optical_depth=0.0076,
                 aerosol_exponent=1.26,
                 ozone_template_path=None,
                 keys={}):
        """
        Creates an atmospheric extinction model based on Buton et al.

        By default looks for the ozone template in 
        saltpath.atmospheric_model_path+"/Buton/ozoneTemplate.fits"
        which should be
        $SALTPATH/atmospheric_model/Buton/ozoneTemplate.fits


        If no parameters are provided, the airmass will be set to 1. and the other parameters to the
        average values given by Buton et al 2013:
        are adapted to mean Mauna-Kea summit conditions.

        The ozone template can be provided by hand, but by default it is assumed to live in
        $SALTPATH/atmospheric_model/Buton/ozoneTemplate.fits

        =================================  =================
        Parameter                          Value ± Error
        =================================  =================
        *Mauna Kea*
        ----------------------------------------------------
        Pressure                           616 ± 2 mbar
        Ozone column                       257 ± 23 DU
        Aerosols optical depth @ 1 micron  0.0076 ± 0.0014
        Aerosols angstrom exponent         1.26 ± 1.33
        =================================  =================

        The parameters are stored in a fitparemeters.FitParameters() instance
        
        The ozone template is read in a Func1D() stored in self.ozone_template

        There is a keys dictionnary member, {} by default, to store metadata

        Examples:
        >>> mauna_kea_average_atmosphere = atmosphericmodel.Buton()
        >>> mauna_kea_average_atmosphere.parameter
        pressure: array([ 616.])
        ozone_column: array([ 257.])
        airmass: array([ 1.])
        aerosol_optical_depth: array([ 0.0076])
        aerosol_exponent: array([ 1.26])
        
        >>> wavelength = numpy.linspace(3000., 10000., 200)
        >>> pylab.plot(wavelength, mauna_kea_average_atmosphere(wavelength))
        >>> mauna_kea_average_atmosphere.parameter
        >>> mauna_kea_average_atmosphere.parameter["airmass"] = 2.
        >>> mauna_kea_average_atmosphere.parameter
        pressure: array([ 616.])
        ozone_column: array([ 257.])
        airmass: array([ 2.])
        aerosol_optical_depth: array([ 0.0076])
        aerosol_exponent: array([ 1.26])

        >>> pylab.plot(wavelength, mauna_kea_average_atmosphere(wavelength))


        To plot single components:
        >>> x = np.linspace(2500., 10000., 100)
        >>> plt.plot(x, b(x), "r-")
        >>> plt.plot(x, b.rayleigh_transmission(x), "g-")
        >>> plt.plot(x, b.ozone_transmission(x), "b-")
        >>> plt.plot(x, b.aerosol_transmission(x), "k-")
        >>> p = b.parameter.copy()
        >>> p["aerosol_exponent"] = 3.
        >>> plt.plot(x, b.aerosol_transmission(x, p=p), "k--")

        Changing parameters:
        >>> p = b.parameter
        >>> p["pressure"] = 500.
        This changes what b(x) will return

        To get the parameters:
        >>> 
        """
        self.parameter = fitparameters.FitParameters(["airmass",
                                                      "pressure",
                                                      "ozone_column",
                                                      "aerosol_optical_depth",
                                                      "aerosol_exponent"])

        self.parameter["airmass"][0] = airmass
        self.parameter["pressure"][0] = pressure
        self.parameter["ozone_column"][0] = ozone_column
        self.parameter["aerosol_optical_depth"][0] = aerosol_optical_depth
        self.parameter["aerosol_exponent"][0] = aerosol_exponent

        # By default assumes that the ozone template is in snfit_data
        if ozone_template_path is None:
            ozone_template_path = saltpath.atmospheric_model_path + "/Buton/ozoneTemplate.fits"
        self.read_ozone_template(ozone_template_path)

        self.keys = keys
        
    def get_parameter(self):
        """
        Returns a copy of the parameters

        >>> p = b.parameter
        >>> p["pressure"] = 500.
        Changes b(x)

        >>> p = b.get_parameter()
        >>> p["pressure"] = 500.
        Doesn't change b(x)
        To get the new b(x) do
        >>> b(x, parameter=p)

        NB: you could also do b.parameter.copy()
        """        
        return self.parameter.copy()
    
    def func1d_rayleigh(self, parameter=None):
        """
        pressure is in [mbar]

        sets self.rayleigh to be  Rayleigh extinction in [mag/airmass]
        as a Func1D that will change when self.pressure changes
        """

        if parameter is None:
            parameter = self.parameter

        return parameter["pressure"].full[0] * interpolation.Func1D(buton_base_rayleigh)
                
    def read_ozone_template(self, ozone_template_path):
        """
        Reads the ozone template and puts it in 
        self.ozone_template as a Func1D instance
        """
        # Read wavelength and transmission columns
        ext = 1
        colLbda = "LAMBDA"
        colTrans = "OZONE"
        
        ffile = pyfits.open(ozone_template_path)
        x = ffile[ext].data.field(colLbda)   # Wavelength
        y = ffile[ext].data.field(colTrans)  # Transmission    
        refO3col = ffile[ext].header["REFO3COL"]
        ffile.close()

        self.ozone_template_path = ozone_template_path
        self.ozone_template = interpolation.Func1D(x, numpy.absolute(-2.5 * numpy.log10(y)) / refO3col)
        
    def func1d_ozone(self, parameter=None):
        """
        Sets self.ozone to be the Func1D that calculates the ozone component of the atmospheric transmission
        """
        if parameter is None:
            parameter = self.parameter
        return parameter["ozone_column"].full[0] * self.ozone_template

    def func1d_transmission(self, transmission, parameter=None):
        """
        Returns a func1d that calculates the transmission corresponding to an extinction func1d
        """
        if parameter is None:
            parameter = self.parameter
        return 10 ** (-0.4 * (parameter["airmass"].full[0] * transmission))
    
    def func1d_aerosol(self, parameter=None):
        """
        Sets self.aerosol to be the Func1D that calculates the aerosol component of the atmospheric transmission
        """
        EXT2OPT = .92103403719761834  # LOG10/2.5 = Extinction to opt. thickness
        wavelength_reference = 1.e4
        if parameter is None:
            parameter = self.parameter

        return interpolation.Func1D(lambda x:
                                             parameter["aerosol_optical_depth"].full[0] / EXT2OPT *
                                             (x / wavelength_reference) **
                                             - parameter["aerosol_exponent"].full[0])

    def func1d_atmospheric_extinction(self, parameter=None):
        """
        Returns the Func1D that calculates the atmospheric extiction in [mag/airmass]       
        """
        if parameter is None:
            parameter = self.parameter
            
        return self.func1d_rayleigh(parameter=parameter) + self.func1d_ozone(parameter=parameter) + self.func1d_aerosol(parameter=parameter)

    
    def func1d_atmospheric_transmission(self, parameter=None):
        """
        Returns the atmospheric transmission curve for a given airmass in photons     
        See Buton et al eq.5

        call self(x) to get its values along axis x
        """
        atmospheric_extinction = self.func1d_atmospheric_extinction(parameter=parameter)
        if parameter is None:
            parameter = self.parameter
        return 10 ** (-0.4 * (parameter["airmass"].full[0] * atmospheric_extinction))

    def rayleigh_extinction(self, x, parameter=None, jac=False):
        """
        Convenience function to plot the raileght component of the atmospheric
        extinction corresponding to the parameters of the Buton instance.
        
        If parameter is set, will plot with these parameters instead.
        """
        rayleigh = self.func1d_rayleigh(parameter=parameter)

        if jac:
            raise NotImplemented("Jacobian calculation not implemented")
        else:
            return rayleigh(x)

    def ozone_extinction(self, x, parameter=None, jac=False):
        """
        Convenience function to plot the ozone component of the atmospheric
        extinction corresponding to the parameters of the Buton instance.
        
        If p is set, will plot with these parameters instead.
        """
        ozone = self.func1d_ozone(parameter=parameter)

        if jac:
            raise NotImplemented("Jacobian calculation not implemented")
        else:
            return ozone(x)

    def aerosol_extinction(self, x, parameter=None, jac=False):
        """
        Convenience function to plot the aerosol component of the atmospheric
        extinction corresponding to the parameters of the Buton instance.
        
        If p is set, will plot with these parameters instead.
        """
        aerosol = self.func1d_aerosol(parameter=parameter)

        if jac:
            raise NotImplemented("Jacobian calculation not implemented")
        else:
            return aerosol(x)

    def atmospheric_extinction(self, x, parameter=None, jac=False):
        """
        A convenience function to be able to plot the atmospheric extinction corresponding to the parameters
        of the Buton instance
        """
        atmospheric_extinction = self.func1d_atmospheric_extinction(parameter=parameter)

        if jac:
            raise NotImplemented("Jacobian calculation not implemented")
        else:
            return atmospheric_extinction(x)

        
    def rayleigh_transmission(self, x, parameter=None):
        """
        Convenience function to plot the raileght component of the atmospheric
        transmission corresponding to the parameters of the Buton instance.
        
        If parameter is set, will plot with these parameters instead.
        """
        rayleigh = self.func1d_rayleigh(parameter=parameter)

        if parameter is None:
            parameter = self.parameter
        return 10 ** (-0.4 * (parameter["airmass"].full[0] * rayleigh(x)))

    def ozone_transmission(self, x, parameter=None):
        """
        Convenience function to plot the raileght component of the atmospheric
        transmission corresponding to the parameters of the Buton instance.
        
        If parameter is set, will plot with these parameters instead.
        """
        ozone = self.func1d_ozone(parameter=parameter)

        if parameter is None:
            parameter = self.parameter
        return 10 ** (-0.4 * (parameter["airmass"].full[0] * ozone(x)))


    def aerosol_transmission(self, x, parameter=None):
        """
        Convenience function to plot the raileght component of the atmospheric
        transmission corresponding to the parameters of the Buton instance.
        
        If p is set, will plot with these parameters instead.
        """
        aerosol = self.func1d_aerosol(parameter=parameter)

        if parameter is None:
            parameter = self.parameter
        return 10 ** (-0.4 * (parameter["airmass"].full[0] * aerosol(x)))
                          
        
    def __call__(self, x, parameter=None, jac=False):
        atmospheric_transmission = self.func1d_atmospheric_transmission(parameter=parameter)

        if jac:
            raise NotImplemented("Jacobian calculation not implemented")
        else:
            return atmospheric_transmission(x)
        

class Buton_telluric(object):
    def __init__(self, airmass=1.,
                 H2Ow_power=0.6,
                 H2Os_power=0.6,
                 O2_power=0.58,
                 H2Ow_intensity=1.,
                 H2Os_intensity=1.,
                 O2_intensity=1.,                 
                 telluric_template_path=None,
                 keys={}):
        """Creates a telluric line absorption model based on Buton et al.

        By default looks for the telluric template in 
        saltpath.atmospheric_model_path+"/Buton/SNIFS_extinction.fits"
        which should be
        $SALTPATH/atmospheric_model/Buton/SNIFS_extinction.fits

        The telluric data is in the 2nd extension. Note that the first extension
        contains the average Mauna Kea extinction as implemented in class Buton.
        
        If no parameters are provided, the airmass will be set to 1. and the other parameters to the
        average values given by Buton et al 2013.

        The telluric template can be provided by hand

        The parameters are stored in a fitparameters.FitParameters() instance
        
        The telluric template is read in 3 different Func1D() stored in 
        self.H2Ow_template
        self.H2Os_template
        self.O2_template
        
        Those are telluric extinctions in mmags
        There is a keys dictionnary member, {} by default, to store metadata        

        Examples:
        >>> mauna_kea_telluric = atmosphericmodel.Buton_telluric()
        >>> mauna_kea_telluric.parameter
        H2Os_power: array([ 0.6])
        O2_intensity: array([ 1.])
        airmass: array([ 1.])
        O2_power: array([ 0.58])
        H2Ow_intensity: array([ 1.])
        H2Os_intensity: array([ 1.])
        H2Ow_power: array([ 0.6])

        >>> wavelength = numpy.linspace(3000., 10000., 200)
        >>> pylab.plot(wavelength, mauna_kea_telluric(wavelength))
        >>> H2Ow_transmission = mauna_kea_telluric.func1d_H2Ow_transmission() # This function will NOT change when mauna_kea_telluric.parameter is changed
        >>> pylab.plot(wavelength, H2Ow_transmission(wavelength))

        >>> pylab.plot(wavelength, mauna_kea_telluric.H2Ow_transmission(wavelength))
        >>> mauna_kea_telluric.parameter["H2Ow_intensity"] = 2.
        >>> pylab.plot(wavelength, mauna_kea_telluric.H2Ow_transmission(wavelength))

        """
        self.parameter = fitparameters.FitParameters(["airmass",
                                                      "H2Ow_power",
                                                      "H2Os_power",
                                                      "O2_power",
                                                      "H2Ow_intensity",
                                                      "H2Os_intensity",
                                                      "O2_intensity"])        

        self.parameter["airmass"][0] = airmass
        self.parameter["H2Ow_power"][0] = H2Ow_power
        self.parameter["H2Os_power"][0] = H2Os_power
        self.parameter["O2_power"][0] = O2_power
        
        self.parameter["H2Ow_intensity"][0] = H2Ow_intensity
        self.parameter["H2Os_intensity"][0] = H2Os_intensity
        self.parameter["O2_intensity"][0] = O2_intensity

        # By default assumes that the telluric template is in snfit_data
        if telluric_template_path is None:
            telluric_template_path = saltpath.atmospheric_model_path + "/Buton/SNIFS_extinction.fits"
        self.read_telluric_template(telluric_template_path)

        self.keys = keys
        
    def get_parameter(self):
        """
        Returns a copy of the parameters
        
        Warning:
        >>> p = b.parameter
        >>> p["pressure"] = 500.
        Changes b(x)

        NB: you could also do b.parameter.copy()
        """
        return self.parameter.copy()
        
    def read_telluric_template(self, telluric_template_path):
        """
        Reads the telluric template and puts it in 
        self.H2Ow_template
        self.H2Os_template
        self.O2_template
        as a Func1D instances

        The templates are extinctions in mmags
        """
        # Read wavelength and transmission columns
        ext = 2
        
        fh = pyfits.open(telluric_template_path)
        wavelength = fh[ext].data.field("LAMBDA")  # Wavelength
        H2Ow = fh[ext].data.field('H2Ow    ')
        H2Os = fh[ext].data.field('H2Os    ')
        O2 = fh[ext].data.field('O2      ') 
        fh.close()

        self.telluric_template_path = telluric_template_path
        self.H2Ow_template = interpolation.Func1D(wavelength, H2Ow)
        self.H2Os_template = interpolation.Func1D(wavelength, H2Os)
        self.O2_template = interpolation.Func1D(wavelength, O2)
        
    def func1d_H2Ow_extinction(self, parameter=None):
        """
        Returns the Func1D that calculates the H2Ow component of the telluric
        extinction accounting for intensity, airmass and power
        
        """
        if parameter is None:
            parameter = self.parameter

        return parameter["airmass"].full[0] ** parameter["H2Ow_power"].full[0] * parameter["H2Ow_intensity"].full[0] * self.H2Ow_template 

    def func1d_H2Ow_transmission(self, parameter=None):
        """
        Returns the transmission of the H2Ow component accounting for the
        airmass and intensity of the component
        """
        H2Ow_extinction = self.func1d_H2Ow_extinction(parameter=parameter)
        return 10 ** (-0.4 * H2Ow_extinction)
    
    def func1d_H2Os_extinction(self, parameter=None):
        """
        Returns the Func1D that calculates the H2Ow component of the telluric
        extinction accounting for intensity, airmass and power
        
        """
        if parameter is None:
            parameter = self.parameter

        return parameter["airmass"].full[0] ** parameter["H2Os_power"].full[0] * parameter["H2Os_intensity"].full[0] * self.H2Os_template 

    def func1d_H2Os_transmission(self, parameter=None):
        """
        Returns the transmission of the H2Os component accounting for the
        airmass and intensity of the component
        """
        H2Os_extinction = self.func1d_H2Os_extinction(parameter=parameter)
        return 10 ** (-0.4 * H2Os_extinction)    
    
    def func1d_O2_extinction(self, parameter=None):
        """
        Returns the Func1D that calculates the H2Ow component of the telluric
        extinction accounting for intensity, airmass and power
        
        """
        if parameter is None:
            parameter = self.parameter

        return parameter["airmass"].full[0] ** parameter["O2_power"].full[0] * parameter["O2_intensity"].full[0] * self.O2_template 
    
    def func1d_O2_transmission(self, parameter=None):
        """
        Returns the transmission of the O2 component accounting for the
        airmass and intensity of the component
        """
        O2_extinction = self.func1d_O2_extinction(parameter=parameter)
        return 10 ** (-0.4 * O2_extinction)

    def func1d_telluric_extinction(self, parameter=None):
        """
        Returns the Func1D that calculates the telluric extiction 
        
        Note that since each component has a different airmass power, the
        airmass dependence is already included.  
        """
        extinction = self.func1d_H2Ow_extinction(parameter=parameter) + \
                     self.func1d_H2Os_extinction(parameter=parameter) + \
                     self.func1d_O2_extinction(parameter=parameter)
        
        return extinction
    
    def func1d_telluric_transmission(self, parameter=None):
        """
        Returns the full telluric transmission curve including airmass
        dependence and relative intensity of the components     
        """
        telluric_extinction = self.func1d_telluric_extinction(parameter=parameter)
        return 10 ** (-0.4 * telluric_extinction)

    def H2Ow_transmission(self, x, parameter=None, jac=False):
        """
        Convenience function that evaluates the H20w transmission 
       
        Also usable to fit only the H2Ow component
        """
        H2Ow_transmission = self.func1d_H2Ow_transmission(parameter=parameter)

        if jac:
            raise NotImplemented("Jacobian calculation not implemented")
        else:
            return H2Ow_transmission(x)

    def H2Os_transmission(self, x, parameter=None, jac=False):
        """
        Convenience function that evaluates the H20w transmission 
       
        Also usable to fit only the H2Os component
        """
        H2Os_transmission = self.func1d_H2Os_transmission(parameter=parameter)

        if jac:
            raise NotImplemented("Jacobian calculation not implemented")
        else:
            return H2Os_transmission(x)

    def O2_transmission(self, x, parameter=None, jac=False):
        """
        Convenience function that evaluates the H20w transmission 
       
        Also usable to fit only the O2 component
        """
        O2_transmission = self.func1d_O2_transmission(parameter=parameter)

        if jac:
            raise NotImplemented("Jacobian calculation not implemented")
        else:
            return O2_transmission(x)

    def telluric_extinction(self, x, parameter=None, jac=False):
        """
        Facility method to easily access the full telluric extinction
        """
        telluric_extinction = self.func1d_telluric_extinction(parameter=parameter)

        if jac:
            raise NotImplemented("Jacobian calculation not implemented")
        else:
            return telluric_extinction(x)

        
    def __call__(self, x, parameter=None, jac=False):
        telluric_transmission = self.func1d_telluric_transmission(parameter=parameter)

        if jac:
            raise NotImplemented("Jacobian calculation not implemented")
        else:
            return telluric_transmission(x)



class Buton(Buton_atmosphere, Buton_telluric):
    def __init__(self, airmass=1.,
                 pressure=616.,
                 ozone_column=257.,
                 aerosol_optical_depth=0.0076,
                 aerosol_exponent=1.26,
                 ozone_template_path=None,
                 H2Ow_power=0.6,
                 H2Os_power=0.6,
                 O2_power=0.58,
                 H2Ow_intensity=1.,
                 H2Os_intensity=1.,
                 O2_intensity=1.,                 
                 telluric_template_path=None,
                 keys={}):
        """
        This is an object whose purpose is to provide a way to handle simultaneously
        Buton_atmosphere and Buton_telluric into a single instance.
        
        """
        
        # Initialize the parameters
        self.parameter = fitparameters.FitParameters(["airmass",
                                                      "pressure",
                                                      "ozone_column",
                                                      "aerosol_optical_depth",
                                                      "aerosol_exponent",
                                                      "H2Ow_power",
                                                      "H2Os_power",
                                                      "O2_power",
                                                      "H2Ow_intensity",
                                                      "H2Os_intensity",
                                                      "O2_intensity"])        
        
        self.parameter["airmass"][0] = airmass
        self.parameter["pressure"][0] = pressure
        self.parameter["ozone_column"][0] = ozone_column
        self.parameter["aerosol_optical_depth"][0] = aerosol_optical_depth
        self.parameter["aerosol_exponent"][0] = aerosol_exponent
        
        self.parameter["H2Ow_power"][0] = H2Ow_power
        self.parameter["H2Os_power"][0] = H2Os_power
        self.parameter["O2_power"][0] = O2_power
        
        self.parameter["H2Ow_intensity"][0] = H2Ow_intensity
        self.parameter["H2Os_intensity"][0] = H2Os_intensity
        self.parameter["O2_intensity"][0] = O2_intensity

        
        # Create the two components of the model and have them use the same
        # parameter vector
        self.atmosphere = Buton_atmosphere()
        self.atmosphere.parameter = self.parameter
        self.ozone_template_path = self.atmosphere.ozone_template_path
        self.ozone_template = self.atmosphere.ozone_template
         
        
        self.telluric = Buton_telluric()
        self.telluric.parameter = self.parameter
        self.telluric_template_path = self.telluric.telluric_template_path
        self.H2Ow_template = self.telluric.H2Ow_template
        self.H2Os_template = self.telluric.H2Os_template
        self.O2_template = self.telluric.O2_template

        self.keys = keys

    def get_parameter(self):
        """
        Returns a copy of the parameters

        Warning: p = b.parameter 
        Changing p will change what b(x) returns

        NB: you could also do b.parameter.copy()
        """
        return self.parameter.copy()
        

    def func1d_total_transmission(self, parameter=None):
        """
        Returns the atmospheric transmission curve for a given airmass in photons     
        See Buton et al eq.5
        
        And including the telluric lines

        call self(x) to get its values along axis x
        """
        telluric_transmission = self.func1d_telluric_transmission(parameter=parameter)
        atmospheric_transmission = self.func1d_atmospheric_transmission(parameter=parameter)

        total_transmission = telluric_transmission * atmospheric_transmission
        return total_transmission

        
    def __call__(self, x, parameter=None, jac=False):
        """
        Evaluates the total transmission (atmospheric + telluric) on the
        wavelength grid x
        """
        
        total_transmission = self.func1d_total_transmission(parameter=parameter)
        
        if jac:
            raise NotImplemented("Jacobian calculation not implemented")
        else:
            return total_transmission(x)
        
    



r"""
The purpose of regularization is to constrain the surfaces :math:`M_0` and :math:`M_1` in areas of the plane
that are empty of data.
We use a classical regularization scheme, called Generalized Tikhonov Regularization,
consisting in adding a small quadratic penalty.
This penalty is designed to limit the abrupt variations of the model (oscillations)
and to make the model tend slowly towards zero in the absence of data,
in particular at the phases :math:`p < -20` and :math:`p > 60` days.

The :math:`\chi^2_{reg}` is defined as :

.. math::
    \chi^2_{reg} = \mu_{reg} \beta^T P \beta

With :math:`P`, listing parameters in four blocs
([:math:`X_0`, :math:`t_{max}^{B^\star}`, :math:`X_1`, :math:`c`], :math:`M_0`, :math:`M_1`,
[:math:`CL`, :math:`s` , :math:`\eta`, :math:`\kappa`]), by:

.. math::
    \begin{equation}
    P = \left (
    \begin{array}{c c c c}
    0 & 0 & 0 & 0 \\
    0 & P_{\theta_{0|1}} & 0 & 0\\
    0 & 0 & P_{\theta_{0|1}} & 0\\
    0 & 0 & 0 & 0
    \end{array}
    \right)
    \end{equation}

The matrix :math:`P_{\theta_{0|1}}` is defined in the different regularization class.
"""

import numpy as np
from scipy.sparse import coo_matrix, dia_matrix, csc_matrix
    

class SplineRegul2D(object):
    r"""
    This regularization focuses on the model variations in phase.
    The density of spectra allows the basis to be constrained correctly, and the density of points
    on the spectra seems to limit the oscillation phenomena.

    The :math:`\chi^2_{reg}` is defined as :

    .. math::
        \chi^2_{reg} &= \mu_{reg} \left[\sum_{i=0}^{N_{\lambda}-1} \sum_{i=0}^{N_{phase}-1}
        \left(\theta_{0|1}^{j, i+1} - \theta_{0|1}^{j, i}\right)^2 + (\theta_{0|1}^{ji})^2 \right] \\

    This gives that by default :math:`P_{\theta_{0|1}}` is given by:

    .. math::
        P_{\theta_{0|1}} = \begin{pmatrix}
        2 & -1 & 0 &  \cdots & & && 0 \\
        -1 & 3 & -1 & \cdots & &  &&0 \\
        \cdots & & & \cdots & & & &\cdots\\
        \cdots & & -1 & 3 & -1 & & &\cdots\\
        \cdots & & & \cdots & & & &\cdots\\
        0 & && \cdots &  &  -1 & 3 & -1 \\
        0 & & &\cdots &  &  & -1 & 2 \\
        \end{pmatrix}

    else :math:`P_{\theta_{0|1}}` is only the identity matrix.

    Attributes
    ----------
    pars : numpy.array
        Model parameter
    mu : float
        Amplitude of the regularization penalty
    matrix_free : scipy.sparse.coo_matrix
        :math:`P_{\theta_{0|1}}` of only the free parameters.
    matrix_full : scipy.sparse.coo_matrix
        :math:`P_{\theta_{0|1}}` of all the parameters.
    """

    def __init__(self, model, surfaces_reg=['M0'], mu=1.E-6, deriv=True):
        """
        Constructor.
        Creation of the :math:`P` matrix.

        Parameters
        ----------
        model : nacl.salt
            Model to regularize.
        surfaces_reg : list
            List of surfaces to regularized
        mu : float
            Amplitude of the contribution to the :math:`\chi^2`
        deriv : bool
            if True :math:`P_{\theta_{0|1}}` is used.
            Else its a diagonal matrix.
        """
        self.pars = model.pars.copy()
        self.mu = mu

        # build the regularization matrix for the spline parameters
        n = len(self.pars['M0'].full)
        if deriv:
            data = np.vstack(([3.] * n, [-1.] * n, [-1.] * n))
            data[0, 0] = data[0, -1] = 2.
            idx_data = [0, -1, 1]
        else:
            # only diagonal term
            data = np.ones(n).astype(float)
            idx_data = [0]
            
        matrix = dia_matrix((data, idx_data), shape=(n, n)).tocoo()
        # embed it in a larger matrix, that corresponds to the model parameters
        n_pars = len(self.pars.free)
        i = self.pars['M0'].indexof(matrix.row)
        j = self.pars['M0'].indexof(matrix.col)
        idx = (i >= 0) & (j >= 0)
        matrix_free0 = coo_matrix((matrix.data[idx]*self.mu, (i[idx], j[idx])), shape=(n_pars, n_pars))

        if 'M1' in surfaces_reg:
            n1 = len(self.pars['M1'].full)
            if deriv:
                data1 = np.vstack(([3.] * n1, [-1.] * n1, [-1.] * n1))
                data1[0, 0] = data1[0, -1] = 2.
                idx_data1 = [0, -1, 1]
            else:
                # only diagonal term
                data1 = np.ones(n1).astype(float)
                idx_data1 = [0]
                
            matrix_1 = dia_matrix((data1, idx_data1), shape=(n1, n1)).tocoo()
            i1 = self.pars['M1'].indexof(matrix_1.row)
            j1 = self.pars['M1'].indexof(matrix_1.col)
            idx1 = (i1 >= 0) & (j1 >= 0)
            matrix_free1 = coo_matrix((matrix_1.data[idx1]*self.mu, (i1[idx1], j1[idx1])), shape=(n_pars, n_pars))
            self.matrix_free = (matrix_free0 + matrix_free1)
        else:
            self.matrix_free = matrix_free0

        pp = self.pars.copy()
        pp.release()
        n_pars = len(pp.full)
        i = pp['M0'].indexof(matrix.row)
        j = pp['M0'].indexof(matrix.col)

        if 'M1' in surfaces_reg:
            i1 = pp['M1'].indexof(matrix_1.row)
            j1 = pp['M1'].indexof(matrix_1.col)
            self.matrix_full = coo_matrix((matrix.data*self.mu, (i, j)),
                                          shape=(n_pars, n_pars))+coo_matrix((matrix_1.data * self.mu, (i1, j1)),
                                                                             shape=(n_pars, n_pars))
        else:
            self.matrix_full = coo_matrix((matrix.data*self.mu, (i, j)), shape=(n_pars, n_pars))  # sans les self.mu

    def __call__(self, beta):
        """
        Calculate the regularization contribution to the final :math:`\chi^2`

        Parameters
        ----------
        beta : array
            Model parameters.

        Returns
        -------
        v : float
            Value of the :math:`\chi^2_{reg}`
        matrix_free : scipy.sparse.coo_matrix
            :math:`P_{\theta_{0|1}}`
        """
        self.pars.free = beta
        # there is probably a way to compute this faster
        v = np.dot(self.pars.full.T, self.matrix_full * self.pars.full)
        return v, self.matrix_free


class SplineRegul2DSalt3(object):
    r"""
    This regularization is the one defined in SALT3 (Kenworthy et al. 2021)
    It is focused on the model variations in phase and wavelength.
    The density of spectra allows the basis to be constrained correctly, and the density of points
    on the spectra seems to limit the oscillation phenomena.

    The :math:`\chi^2_{reg}` is defined as :

    .. math::
        \chi^2_{reg} = \mu_{reg} \sum_{j=0}^{2N_{\lambda}} \sum_{i=0}^{2N_{phase}} \left[\frac{\left(
        \left.\frac{\partial M_{0|1}}{\partial p} \right|_{p = p_i,
        \lambda = \lambda_j }\right)^2}{N_{eff}(p_i,\lambda_j)} +
        \frac{\left( \left.\frac{\partial M_{0|1}}{\partial \lambda}
        \right|_{p = p_i, \lambda = \lambda_j } \right)^2}{N_{eff}(p_i,\lambda_j)}
        + \frac{1}{N_{eff}(p_i,\lambda_j)}\left( \frac{\partial M_{0|1}}{\partial p}
        \times \frac{\partial M_{0|1}}{\partial \lambda} + M_{0|1}\frac{\partial^2 M_{0|1}}{\partial
        \lambda \partial p} \right) \right]
    """
    def __init__(self, model, surfaces_reg=['M0'], mu=1.E-6):
        r"""
        Constructor.
        Creation of the :math:`P` matrix.

        Parameters
        ----------
        model : nacl.salt
            Model to regularize.
        surfaces_reg : list
            List of surfaces to regularized
        mu : float
            Amplitude of the contribution to the :math:`\chi^2`
        """
        self.model = model
        self.pars = model.pars.copy()
        self.mu = mu

        self.ph1_range, self.wl1_range, self.N_eff = None, None, None

        self.surfaces_reg = surfaces_reg
        
        self.wl_range = np.linspace(self.model.wl_range[0], self.model.wl_range[1],
                                    self.model.basis.bx.nj) 
        self.ph_range = np.linspace(self.model.phase_range[0], self.model.phase_range[1],
                                    self.model.basis.by.nj)
        
        self.SP = self.model.training_dataset.spec_data
        # build the regularization matrix for the spline parameters                                     

    def calculate_n_eff(self):
        """
        Calculate the number of data point per bin of wavelength and phase.
        """
        yy_sp, xx_sp = [], []
        bins = self.wl_range
        for sp in self.model.training_dataset.spectra:
            x_sp = sp.data['Wavelength']/(1+sp.data['ZHelio'])
            y_sp = (sp.data['Date']-self.pars['tmax'].free[sp.data['sn_id']])/(1+sp.data['ZHelio'])
            idx_bins = np.where((x_sp.min() < bins) & (x_sp.max() > bins))
            xx_sp.append(bins[idx_bins])
            yy_sp.append(y_sp[0] * np.ones_like(idx_bins[0]))

        xx_sp = np.hstack(xx_sp)
        yy_sp = np.hstack(yy_sp)

        self.ph1_range = np.hstack((self.ph_range, self.ph_range[-1]+1))
        self.wl1_range = np.hstack((self.wl_range, self.wl_range[-1]+1))

        self.N_eff, _, _ = np.histogram2d(xx_sp, yy_sp, bins=[self.wl1_range, self.ph1_range])
        self.N_eff += 1

    def __call__(self, beta):
        """
        Calculate the regularization contribution to the final :math:`\chi^2`

        Parameters
        ----------
        beta : array
            Model parameters.

        Returns
        -------
        v : float
            Value of the :math:`\chi^2_{reg}`
        matrix_free : scipy.sparse.coo_matrix
            :math:`P_{\theta_{0|1}}`
        """
        self.pars.free = beta
        self.calculate_n_eff()

        x, y = np.meshgrid(self.wl_range, self.ph_range)
        model_eval = self.model.basis.eval(x.ravel(), y.ravel())
        jacobian_model_eval_wavelength, jacobian_model_eval_phase = self.model.basis.gradient(x.ravel(), y.ravel())

        _, _, hessian_model_eval_phase = self.model.basis.hessian(x.ravel(), y.ravel())
        
        ddiag = jacobian_model_eval_wavelength * jacobian_model_eval_phase - model_eval * hessian_model_eval_phase
        
        jacobian_model_eval_phase.data *= jacobian_model_eval_phase.data * self.mu * 0.75/1000  # diff or not
        jacobian_model_eval_wavelength.data *= jacobian_model_eval_wavelength.data * self.mu
        jacobian_model_eval_phase.data /= self.N_eff.ravel()[jacobian_model_eval_phase.row]
        jacobian_model_eval_wavelength.data /= self.N_eff.ravel()[jacobian_model_eval_wavelength.row]

        ddiag.data *= self.mu/self.N_eff.ravel()[ddiag.tocoo().row]  # diff or not
        
        matrix = (jacobian_model_eval_phase + jacobian_model_eval_wavelength + ddiag)
        matrix = matrix.tocoo()
        n_pars = len(self.pars.free)
        matrix_free = coo_matrix((np.zeros_like(matrix.data), (matrix.row, matrix.col)), shape=(n_pars, n_pars))

        if 'M0' in self.surfaces_reg:
            i = self.pars['M0'].indexof(matrix.row)
            j = self.pars['M0'].indexof(matrix.col)
            idx = (i >= 0) & (j >= 0)
            matrix_free0 = coo_matrix((matrix.data[idx], (i[idx], j[idx])), shape=(n_pars, n_pars))
            matrix_free += matrix_free0

        if 'M1' in self.surfaces_reg:
            # n1 = len(self.pars['M1'].full)
            i1 = self.pars['M1'].indexof(matrix.row)
            j1 = self.pars['M1'].indexof(matrix.col)
            idx1 = (i1 >= 0) & (j1 >= 0)
            matrix_free1 = coo_matrix((matrix.data[idx1], (i1[idx1], j1[idx1])), shape=(n_pars, n_pars))
            matrix_free += matrix_free1

        pp = self.pars.copy()
        pp.release()
        n_pars = len(self.pars.full)
        matrix_full = coo_matrix((np.zeros_like(matrix.data), (matrix.row, matrix.col)), shape=(n_pars, n_pars))

        if 'M0' in self.surfaces_reg:
            i = self.pars['M0'].indexof(matrix.row)
            j = self.pars['M0'].indexof(matrix.col)
            idx = (i >= 0) & (j >= 0)
            matrix_full0 = coo_matrix((matrix.data[idx], (i[idx], j[idx])), shape=(n_pars, n_pars))
            matrix_full += matrix_full0

        if 'M1' in self.surfaces_reg:
            # n1 = len(self.pars['M1'].full)
            i1 = self.pars['M1'].indexof(matrix.row)
            j1 = self.pars['M1'].indexof(matrix.col)
            idx1 = (i1 >= 0) & (j1 >= 0)
            matrix_full1 = coo_matrix((matrix.data[idx1], (i1[idx1], j1[idx1])), shape=(n_pars, n_pars))
            matrix_full += matrix_full1

        self.matrix_full = matrix_full
        self.matrix_free = matrix_free

        v = np.dot(self.pars.full.T, self.matrix_full * self.pars.full)
        return v, self.matrix_free



"""
Tools to use snfit_data
"""

import os

SALTPATH = os.getenv("SALTPATH")

fitmodel_filename = os.path.join(SALTPATH, "fitmodel.card")

atmospheric_model_path = os.path.join(SALTPATH, "atmospheric_model")

stellar_libs_path = os.path.join(SALTPATH, "stellar_libs")

galaxy_template_path = os.path.join(SALTPATH, "galaxy_templates")

sn_template_path = os.path.join(SALTPATH, "sn_templates")

def reload_saltpath(saltpath=os.getenv("SALTPATH")):
    global SALTPATH
    global fitmodel_filename
    global atmospheric_model_path
    global stellar_libs_path
    global sn_template_path
    
    SALTPATH = saltpath

    fitmodel_filename = os.path.join(SALTPATH, "fitmodel.card")

    atmospheric_model_path = os.path.join(SALTPATH, "atmospheric_model")

    stellar_libs_path = os.path.join(SALTPATH, "stellar_libs")

    galaxy_template_path = os.path.join(SALTPATH, "galaxy_templates")

    sn_template_path = os.path.join(SALTPATH, "sn_templates")
    

def read_fitmodel_card(filename=None):
    """
    Reads $SALTPATH/fitmodel.cards or filename if filename is set
    """
    if filename is None:
        return read_card_file(fitmodel_filename)
    else:
        return read_card_file(filename)
        
def read_card_file(filename):
    """ Read a snfit "card" file into a dictionnary.

    Read all entries with the format:
    "@key1 value1
     @key2 ..."

    in file "filename.
    RETURN:
    ------
    a dictionnary: {key1: value1, key2: ...}
    """
    cards = {}
    fid = open(filename)
    lines = fid.readlines()
    fid.close()
    for line in lines:
        if not line.startswith('@'):
            continue
        else:
            key = line.split()
            value = " ".join(key[1:])
            key = key[0]
            cards[key[1:]] = value
    return cards




"""
This module contains the Newton Raphson minimizer use in NaCl.
It is a minimizer of a log likelihood function.
This function is composed of different pieces :
 - a non-linear model;
 - a set of contains;
 - regularization;
 - variance models : Error Snake and Color Scatter;
 - calibration propagation term.
"""


import numpy as np
import scipy.sparse
from scipy.sparse import coo_matrix, csc_matrix
from sksparse import cholmod
import time
import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.DEBUG)


class Minimizer(object):
    r"""
    Newton Raphson minimizer.

    Iterative algorithm that calculates for each iteration the increment :math:`\delta \beta`: 

    .. math :: \delta \beta = -H^{-1} \nabla \chi^2

    where :math:`\chi^2` is :math:`-2 \log(L)` and implemented in the Chi2Model class, 
    :math:`\nabla \chi^2` is the gradient and :math:`H` is the hessian.

    Attributes
    ----------
    pull_function : ModelPulls
        Class of pull evaluation, difference of model and data, on variance model.
        If no variance model is considered, the denominator is the 'FluxErr' of the data.
    constrains : Constraints2D
        Class of constraints, evaluating the constraint as a function of the free parameters of the model.
    regularization : SplineRegul2D
        Class of regularization term, constraints splines parameters of surfaces without data.
    log : list
        Debug to keep in memory at each iteration all useful information,
        :math:`\beta, d\beta, \gamma, d\gamma, \chi^2(\beta), \chi^2(\beta+d\beta)`
    beta_cholmod : float
        Term in order to get the hessian positive definite (odg :math:`\sim 10^{-10}`)
    N_ITER : int
        Number of iteration maximum, after which the fit stops.
    DF_STOP :
        Threshold of :math:`\Delta \chi^2` that defines the convergence.
    lambda_c : array
        Mean rest-frame wavelength of each Light Curve.
    lambda_c_red :
        Reduced mean rest-frame wavelength of each Light Curve.
    beta : array
        Iteration model parameters.
    gamma : array
        Iteration variance model parameters.
    dbeta : array
        Iteration increment on model parameters.
    dgamma : array
        Iteration increment on variance model parameters.
    dsigma_kappa : array
        Increment on color scatter parameters.
    decrement : float
        Iteration :math:`\Delta \chi^2`
    hess : coo_matrix
        Iteration hessian matrix of the :math:`\chi^2`
    grad : array
        Iteration gradient of the :math:`\chi^2`
    increment : array
        increment
    """
    
    def __init__(self, pull_function,
                 constrains=None, regularization=None, color_scatter = None,
                 n_iter=1000,
                 df_stop=1e-5, log=None, beta_cholmod=1e-10):
        r"""
        Initiate the Minimizer class : 

        Parameters
        ----------
        pull_function : ModelPulls
            Class of pull evaluation, difference of model and data, on variance model.
            If no variance model is considered, the denominator is the 'FluxErr' of the data.
        constrains : Constraints2D
            Class of constraints, evaluating the constraint as a function of the free
            parameters of the model.
        regularization : SplineRegul2D
            Class of regularization term, constraint splines parameters of surfaces without data.
        color_scatter :  ColorScatter
            Color scatter function that evaluate the color scatter and its derivatives.
        n_iter : int
            Number of iteration maximum, after which the fit stops. 
        df_stop : float 
            Threshold of :math:`\Delta \chi^2` that defines the convergence.
        log : bool
            Debug to keep in memory at each iteration all useful information,
            :math:`\beta, d\beta, \gamma, d\gamma, \chi^2(\beta), \chi^2(\beta+d\beta)`
        beta_cholmod : float
            Term in order to get the hessian positive definite (odg :math:`\sim 10^{-10}`)
        """
        self.pull_function = pull_function
        self.constrains = constrains
        self.regularization = regularization
        self.color_scatter = color_scatter
        self.log = log

        self.beta_cholmod = beta_cholmod
        self.N_ITER = n_iter
        self.DF_STOP = df_stop

        if self.log is not None:
            self.log = []

        self.lambda_c = self.pull_function.model.lambda_c
        self.lambda_c_red = self.pull_function.model.lambda_c_red
        self.beta, self.gamma, self.dbeta, self.dgamma, self.dsigma_kappa = None, None, None, None, None
        self.decrement = None
        self.hess, self.grad, self.increment = None, None, None

    def log_append(self, iteration, chi2_1, f_val, detail1, detail2):
        r"""
        Add to log the information of the iteration.
        
        Parameters
        ----------
        iteration : int
            Iteration number.
        chi2_1 : float
            :math:`\chi^2(\beta_{iteration - 1})`
        f_val : float
            :math:`\chi^2(\beta_{iteration})`
        detail1 : dict
            Values of the different component of the  :math:`\chi^2`, of the iteration number : iteration - 1
        detail2 : dict
            Values of the different component of the  :math:`\chi^2` of the iteration number : iteration
        """
        if self.log is not None:
            if self.sigma_kappa_fit & (self.sigma_kappa is not None):
                self.log.append((iteration, self.beta, self.gamma, self.dgamma, self.dbeta,
                                 chi2_1, f_val, self.M, detail1, detail2,
                                 self.sigma_kappa.copy(), self.dsigma_kappa))
            else:
                self.log.append((iteration, self.beta, self.gamma, self.dgamma, self.dbeta,
                                 chi2_1, f_val, self.M, detail1, detail2, None, None))
        
    def parameter_update(self):
        r"""
        Updated the values of the model parameters :math:`\beta` and the variances model parameters : :math:`\gamma`,
        :math:`\sigma_\kappa` for the considered iteration.
        
        Note
        ----
        In increment the parameters are stored as following :  :math:`(\beta, \gamma, \sigma_\kappa)`
        """
        if self.sigma_kappa_fit & (self.sigma_kappa is not None):
            # Slice the sigma_kappa parameters in increment
            self.dsigma_kappa = self.increment[self.fchi2.N_PAR:]
            # Slice the gamma parameters in increment
            self.dgamma = self.increment[len(self.beta): - len(self.dsigma_kappa)]
        else:
            self.dgamma = self.increment[len(self.beta):]

        self.dbeta = self.increment[: len(self.beta)]
        t = 1.
        
        self.beta = self.beta + t * self.dbeta
        
        if len(self.pull_function.VarianceModel.pars.free) != 0:        
            self.gamma = self.gamma + t * self.dgamma
        if self.sigma_kappa_fit & (self.sigma_kappa is not None):
            self.sigma_kappa += t * self.dsigma_kappa
            print('sigma_kappa : ', self.sigma_kappa)
            print('dsigma_kappa : ', self.dsigma_kappa)

    def print_end_iter(self, t, chi2_1, f_val, n_dof, detail1, detail2):
        r"""
        Print the iteration information of the :math:`\chi^2`.
        
        Parameters
        ----------
        t : float
            Line-search parameters, default value is 1
        chi2_1 : float                                                                                  
            :math:`\chi^2(\beta_{iteration - 1})`                                                       
        f_val : float
            :math:`\chi^2(\beta_{iteration})`
        n_dof : float
            Number of degrees of freedom, calculated as the difference of the number
            of data point and the number of free parameters.
        detail1 : dict                       
            Values of the different component of the  :math:`\chi^2`, of the iteration number : iteration - 1
        detail2 : dict
            Values of the different component of the  :math:`\chi^2` of the iteration number : iteration
        """
        print('chi^2 contribution  i-1: ', detail1, '\n', 'chi^2 contribution  i: ',  detail2, '\n')
        if (np.isnan(f_val) ^ np.isinf(f_val)).sum() > 0:
            raise ValueError(f'Value of f_val : {f_val} ')
        self.decrement = chi2_1 - f_val
        logging.info('step size: %g, objective: %g -> %g, decrement: %g, D.o.F.: %d, objective/dof: %g'
                     % (t, chi2_1, f_val, self.decrement, n_dof, f_val/n_dof))
        print(f"t_tot : {self.titer[-1]},  t_cho : {self.tcholesky[-1]}")

    def __call__(self, beta0, sigma_kappa=None, gamma0=None, eta_covmatrix=None,
                 sigma_kappa_fit=True):
        r"""
        Find the parameters of the model and the variance model by minimizing the log likelihood function.
        
        Parameters
        ----------
        beta0 : array
            Initial model's parameters
        sigma_kappa : array or None
            Initial color scatter parameters.
            If None, the color scatter is not fitted.
            Kappa parameters of the model should be fixed, otherwise the degeneracy in the model is not broken.
        gamma0 :  array or None
            Initial error snake.
            If None, the variance of data is given by the square of the error on the measurement.
        eta_covmatrix : array or None
            Calibration zero point incertitude matrix.
            Eta parameters of the model should be fixed, otherwise the degeneracy in the model is not broken.
        sigma_kappa_fit : bool
           True : sigma kappa of fitted
           False : They are not, but if sigma_kappa is not None, kappa parameters are fitted, with a fix color scatter.
        """
        self.gamma0 = self.gamma = gamma0
        self.beta = self.beta0 = beta0.copy()
        self.n = len(beta0)
        
        self.factorCS = None
            
        self.sigma_kappa_fit = sigma_kappa_fit
        self.sigma_kappa = sigma_kappa
        
        if self.gamma0 is None:
            self.gamma = []
                    
        self.M = np.ones(1)
        self.tcholesky, self.tmodel, self.thessian, self.titer = [], [],  [],  []
        for iteration in range(self.N_ITER):
            t0 = time.time()
            print('\n')
            logging.info('iteration: %d' % iteration)
            beta_keep = self.beta.copy()
            if len(self.gamma) != 0:
                gamma_keep = self.gamma.copy()
            hess_keep = self.M.copy()
            
            self.fchi2 = Chi2Model(pull_function=self.pull_function, constrains=self.constrains,
                                   regularization=self.regularization, color_scatter=self.color_scatter,
                                   sigma_kappa_fit=sigma_kappa_fit)
            
            chi2_1, detail1, self.gradf, self.M = self.fchi2(self.beta, self.gamma, self.sigma_kappa,
                                                             eta_covmatrix=eta_covmatrix,
                                                             derivatives=True)
            
            tmod = self.fchi2.model_evaluation
            thess = self.fchi2.hessian_construction
            self.thessian.append(thess)
            self.tmodel.append(tmod)
            
            # solve the normal equation
            t_cho = time.time()            

            # When Error snake is fitted : M not positively definated
            # => should use mode=simplicial (much slower)
            # if len(self.pull_function.VarianceModel.pars.free) == 0:
            #     fact = cholmod.cholesky(self.M.tocsc(), mode='supernodal',
            #                             ordering_method='metis', beta=self.beta_cholmod)
                
            #     print('SUPERNODAL')
            # else :
            fact = cholmod.cholesky(self.M.tocsc(), mode='simplicial',
                                    ordering_method='metis', beta=self.beta_cholmod)
            print('SIMPLICIAL')
            
            self.tcholesky.append(time.time() - t_cho)
            self.increment = fact(-1*self.gradf) 
            
            self.parameter_update()
            
            if self.constrains is not None:
                print('CHECK CONS 1: ', self.constrains(self.beta)*np.sqrt(self.constrains.sig_vals))
                print('Variance X1 : ', self.constrains.var_x1(self.beta)/np.sqrt(self.constrains.mu_pq_var_x1), '\n')
                
            ndof = len(self.pull_function.data) - len(self.beta)
            if len(self.pull_function.VarianceModel.pars.free) != 0:
                ndof -= len(self.gamma)
                
            if self.sigma_kappa_fit & (self.sigma_kappa is not None):
                ndof -= len(self.sigma_kappa) 

            f_val, detail2 = self.fchi2(self.beta, self.gamma, self.sigma_kappa,
                                        eta_covmatrix=eta_covmatrix, derivatives=False)
            t = 1.
            self.titer.append(time.time() - t0)
            self.print_end_iter(t, chi2_1, f_val, ndof, detail1, detail2)
            self.log_append(iteration, chi2_1, f_val, detail1, detail2)
            
            self.iminus1_chi2 = chi2_1
            self.final_chi2 = f_val

            self.iter = iteration
            if self.decrement < 0.0:
                self.beta = beta_keep
                self.M = hess_keep 
                if len(self.gamma) != 0:
                    self.gamma = gamma_keep
                if self.sigma_kappa_fit & (self.sigma_kappa is not None):
                    self.sigma_kappa = self.sigma_kappa - t * self.dsigma_kappa
                self.pull_function.model.pars.free = self.beta
                
                if len(self.pull_function.VarianceModel.pars.free) != 0:
                    self.pull_function.VarianceModel.pars.free = self.gamma

            if self.decrement < self.DF_STOP:
                break


class Chi2Model(object):
    r"""
     compute the full Lagrangian, including 
      - the model,
      - Error snake 
      - Color scatter
      - Constraints [optional]: a quadratic penalty derived from the constraints
      - Regularization [optional]
      - Propagation of calibration:

     .. math:: 
         \chi^{2} = \, & \chi^{2}_{model} + \chi^{2}_{cons} + \chi^{2}_{reg} +
         \chi^{2}_{calib} + \chi^{2}_{CS}  \nonumber \\
         =\, & \ln(|V(\beta, \gamma)|) + R(\beta)^T V^{-1}(\beta, \gamma) R(\beta) \nonumber \\
         + \, &  \mu_{pen} C(\beta)^T C(\beta) \nonumber \\
         +\, & \mu_{reg} \beta^T P \beta \nonumber \\
         + \, & \eta^T V^{-1}_{\eta} \eta \nonumber\\ 
         +\,& \ln{|V_{\kappa}( \sigma_\kappa)|} + \kappa^T V_\kappa^{-1}( \sigma_\kappa) \kappa 


    Attributes
    ----------
    pull_function : ModelPulls
        Class of pull evaluation, difference of model and data, on variance model.
        If no variance model is considered, the denominator is the 'FluxErr' of the data.
    constrains : Constraints2D
        Class of constraints, evaluating the constraint as a function of the free parameters of the model.
    regularization : SplineRegul2D
        Class of regularization term, constraints splines parameters of surfaces without data.
    lambda_c : array
        Mean rest-frame wavelength of each Light Curve.
    lambda_c_red :
        Reduced mean rest-frame wavelength of each Light Curve.
    sigma_kappa_fit : bool 
        Whether color scatter parameters are fitted.
    beta : array
        Model parameters.
    gamma : array
        Variance model parameters.
    hess : coo_matrix
        Hessian matrix of the :math:`\chi^2`
    grad : array
        Gradient of the :math:`\chi^2`
    hessian_construction : float
        Time of model part hessian matrix construction 
    model_evaluation : float
        Time passing evaluating models           
    """    
    def __init__(self, pull_function, constrains=None, 
                 regularization=None, color_scatter=None,
                 sigma_kappa_fit=True):
        """
        Initialization of the class.

        Parameters
        ----------
        pull_function : ModelPulls
            Class of pull evaluation, difference of model and data, on variance model.
            If no variance model is considered, the denominator is the 'FluxErr' of the data.
        constrains : Constraints2D
            Class of constraints, evaluating the constraint as a function of the free parameters of the model.
        regularization : SplineRegul2D
            Class of regularization term, constraints splines parameters of surfaces without data.
        color_scatter :  ColorScatter
            Color scatter function that evaluate the color scatter and its derivatives.
        sigma_kappa_fit : bool
            Whether color scatter parameters are fitted.

        """
        self.pull_function = pull_function
        self.regularization = regularization
        self.constrains = constrains
        self.color_scatter = color_scatter
        
        self.lambda_c_red = self.pull_function.model.lambda_c_red
        self.lambda_c = self.pull_function.model.lambda_c
        self.sigma_kappa_fit = sigma_kappa_fit
        self.hess, self.grad = None, None
        self.hessian_construction, self.model_evaluation = None, None
        
    def __call__(self, beta, gamma, sigma_kappa=None, eta_covmatrix=None, derivatives=False):
        r"""
        Evaluation of the :math:`\chi^2`, with the update parameters
        (math:`\beta`, math:`\gamma`, math:`\sigma_\kappa`).
        Evaluated the gradiant and the hessian if needed.

        Parameters
        ----------
        beta : array
           Model parameters.
        gamma : array
           Error snake parameters.
        sigma_kappa : array
           Color scatter parameters.
        eta_covmatrix : float or array
           Calibration uncertainty matrix.
        derivatives : bool
           If gradient and hessian are needed.

        Returns
        -------
        chi2 : float
             :math:`\chi^2` calculates with the update parameters
        detail : dict
             Component of the :math:`\chi^2`

        if derivatives :
        chi2 : float
             :math:`\chi^2` calculates with the update parameters
        detail : dict
             Component of the :math:`\chi^2`
        grad : array
             Gradient of the :math:`\chi^2`
        hess : scipy.sparse.csc_matrix
             Hessian of the :math:`\chi^2`
        """
        self.detail = {}
        self.beta, self.gamma = beta, gamma
        self.N_PAR = len(beta)
        self.rho = self.beta.copy()
        if len(self.pull_function.VarianceModel.pars.free) != 0:
            self.N_PAR += len(self.gamma)
            self.rho = np.hstack((self.rho, self.gamma))
        self.sigma_kappa = sigma_kappa
        self.chi2 = 0

        self.model_reg(derivatives=derivatives)
        if eta_covmatrix is not None:
            self.calibration_gaussian_prior(eta_covmatrix, derivatives=derivatives)
        if self.constrains is not None:
            self.get_constrains_derivatives(derivatives=derivatives)
        if sigma_kappa is not None:
            self.color_scatter_evaluation(derivatives=derivatives)
        if derivatives:
            return self.chi2, self.detail, self.grad, self.hess
        return self.chi2, self.detail

    def model_reg(self, derivatives=False):
        r"""
        - Compute the model and error snake contribution to the total :math:`\chi^2`,
        grouping all parameters under a single variable :
        
        .. math::
            \rho = \left( \begin{matrix}
            \beta \\
            \gamma \\
            \sigma_\kappa
            \end{matrix}\right)

        We get : 

        .. math::
         \chi^2_{model} = \ln(|V(\rho)|) + R(\rho)^T  V^{-1}(\rho) R(\rho)
        
        For the gradient vector :

        .. math::
            \frac{\partial \chi^2_{model}(\rho)}{\partial \rho_i}  = Tr\left(V^{-1} J_{V,i} \right) -
            2 J_i^T V^{-1} R - R^T V^{-1} J_{V,i} V^{-1} R

        For the hessian matrix :

        .. math::
            H  = \frac{\partial^2 \chi^2_{model}(\rho)}{\partial \rho_i \partial \rho_j}= (H)_{ij} =
            \left( Tr\left( -V^{-1} J_{V,j} V^{-1} J_{V,i} \right) + 2  J_i^T V^{-1} J_j^T\right)_{ij}

        
        - Compute regularization component of the total :math:`\chi^2`:
        
        .. math::
            \chi^2_{reg} = \mu_{reg} \beta^T P \beta

        If derivatives are needed:
        adding to the total gradient :
        
        .. math::
            \frac{\partial \chi^2_{reg}(\beta)}{\partial \beta}  = \mu_{reg} \rho^T P
        
        and to the hessian matrix :
        
        .. math::
            \frac{\partial \chi^2_{reg}(\beta)}{\partial \beta \partial \beta} = 2 \mu_{reg} P        


        Parameters
        ----------
        derivatives : bool
           If gradient and hessian are needed.
        """
        lower_triangular_matrix = None

        tmod = time.time()
        if derivatives:
            # if len(self.gamma) != 0.:
            pull, derivative_pull, val, jacobian, variance, \
                derivative_variance = self.pull_function(self.beta, gamma=self.gamma, jac=True)

        # else:
        # pull, derivative_pull, val, jacobian = self.pull_function(self.beta, jac=True)
        #        variance = self.pull_function.data['FluxErr']**2
        else:
            if len(self.gamma) != 0.:
                pull, val = self.pull_function(self.beta, gamma=self.gamma, jac=False)
            else:
                pull, val = self.pull_function(self.beta, jac=False)
        self.model_evaluation = time.time() - tmod
        res = self.pull_function.data['Flux'] - val
        pull_tilde = pull
        
        if self.regularization is not None:
            _, regularization_matrix = self.regularization(self.beta)
            if regularization_matrix.shape != (len(self.beta), len(self.beta)):
                raise ValueError('wrong regularization matrix dimensions: regularization_matrix.shape=%r'
                                 % regularization_matrix.shape)
            fact = cholmod.cholesky(regularization_matrix.tocsc(), beta=1.E-12)
            lower_triangular_matrix = fact.L()
            pull_tilde = np.hstack((pull, lower_triangular_matrix.T.dot(self.beta)))
        
        self.chi2 += pull_tilde.T.dot(pull_tilde)
        
        self.detail['model'] = pull.T.dot(pull)
        if self.regularization is not None:
            regularization_triangular_matrix = lower_triangular_matrix.T.dot(self.beta)
            self.detail['reg'] = regularization_triangular_matrix.T.dot(regularization_triangular_matrix)

        if len(self.pull_function.VarianceModel.pars.free) != 0:
            variance = self.pull_function.VarianceModel(self.gamma, self.beta)
            self.chi2 += np.log(variance).sum()
            self.detail['var_log_det'] = np.log(variance).sum()
       
        if derivatives:
            if self.regularization is not None:
                derivative_pull_tilde = scipy.sparse.bmat([[derivative_pull, None], [-lower_triangular_matrix.T, None]])
            else:
                derivative_pull_tilde = derivative_pull
            derivative_pull_tilde = derivative_pull_tilde.tocsc()

            t0 = time.time()
            if len(self.pull_function.VarianceModel.pars.free) != 0:
                w_diag = 1/variance
                w_matrix = scipy.sparse.diags(w_diag)
                matrix_variance_pull = (derivative_variance.tocsc().dot(w_matrix.tocsc()))
                
                self.hess = self.hessian_nr(matrix_variance_pull=matrix_variance_pull, derivative_pull=derivative_pull,
                                            derivative_pull_tilde=derivative_pull_tilde)
                self.grad = self.grad_nr(res=res, w_diag=w_diag, matrix_variance_pull=matrix_variance_pull,
                                         pull_tilde=pull_tilde,
                                         derivative_pull_tilde=derivative_pull_tilde)

            else:
                derivative_pull_tildecsr = derivative_pull_tilde.tocsr()
                self.hess = 2 * derivative_pull_tildecsr.T.dot(derivative_pull_tildecsr)
                self.grad = -2. * derivative_pull_tildecsr.T.dot(pull_tilde)
                
            self.hessian_construction = time.time() - t0

    def hessian_nr(self, matrix_variance_pull, derivative_pull, derivative_pull_tilde):
        r"""
        Compute the Hessian of the simultaneous fit of the model parameters
        (beta) and th error snake parameters (gamma):

        .. math::
            H = (H)_{ij} =\left( Tr\left( -V^{-1} J_{V,j} V^{-1} J_{V,i} \right) + 2  J_i^T V^{-1} J_j^T\right)_{ij}

        Parameters
        ----------
        matrix_variance_pull : scipy.sparse.csc_matrix or scipy.sparse.bmat
            Product of the derivative of the error snake wrt to beta and gamma and weight matrix, the variance inverse.
        derivative_pull : array
            Derivative of the pull wrt beta parameters
        derivative_pull_tilde : array
            Concatenation of the derivative of the pull wrt beta
            parameters and the regularization cholesky decomposition term

        Returns
        -------
        Hessian of the model and regularization part :math:`H`
        """
        hessian0 = 2 * derivative_pull_tilde.T.dot(derivative_pull_tilde)
         
        hessian = coo_matrix(hessian0)
        hessian.resize((self.N_PAR, self.N_PAR))
        hessian_csc = hessian.tocsc()

        ddvv = matrix_variance_pull.tocsc()
        trace_derivative_log_det_variance_hessian = -1 * (ddvv.dot(ddvv.T))
        hessian_log_det = trace_derivative_log_det_variance_hessian
        return hessian_csc + hessian_log_det

    def grad_nr(self, res, w_diag, matrix_variance_pull, pull_tilde, derivative_pull_tilde):
        r"""
        Compute the Gradient of the simultaneous fit of the model parameters
        (beta) and th error snake parameters (gamma):
        
        .. math::
            \frac{\partial \chi^2_{model}(\rho)}{\partial \rho_i} = Tr\left( V^{-1} J_{V,i} \right) -
            2 J_i^T V^{-1}  R -  R^T  V^{-1}  J_{V,i}  V^{-1}  R
        
        Parameters 
        ----------
        res : array
           Array of the model residuals.
        w_diag :
           Diagonal of the inverse of the variance matrix.
        matrix_variance_pull : scipy.sparse.csc_matrix or scipy.sparse.bmat
           Product of the derivative of the error snake wrt to beta and gamma and weight matrix, the variance inverse.
        pull_tilde : array 
           Concatenation of the pulls and the regularization cholesky decomposition term
        derivative_pull_tilde : array 
           Concatenation of the derivative of the pull wrt beta
           parameters and the regularization cholesky decomposition term

        Returns
        -------
        Gradient of the model and regularization part :math:`\frac{\partial \chi^2_{model}(\rho)}{\partial \rho_i}`
        """
        grad = -2. * derivative_pull_tilde.T.dot(pull_tilde)
        grad0 = np.concatenate([grad, np.zeros(self.gamma.shape)])
        trace_derivative_log_det_variance_grad = 1 * np.array(matrix_variance_pull.sum(axis=1)).ravel()
        pull_derivation_w_pull = - (matrix_variance_pull.dot((w_diag * res**2)))
        resul = grad0 - trace_derivative_log_det_variance_grad - pull_derivation_w_pull
        return resul
            
    def color_scatter_evaluation(self, derivatives=False):
        r"""
        Evaluation of the color scatter \& its derivatives.
        
        .. math::
            \chi^{2}_{CS}(\kappa, \sigma_\kappa) =  \ln{|V_{\kappa}( \sigma_\kappa)|} +
            \kappa^T V^{-1}_{\kappa}( \sigma_\kappa) \kappa

        adding to de gradient : 

        .. math::
            \frac{\partial \chi^2_{CS}}{\partial \kappa_i} =  2  V_{\kappa}^{-1} \kappa
        
        .. math::
            \frac{\partial \chi^2_{CS}}{\partial \sigma_{\kappa,i}} =  Tr\left(V_{\kappa}^{-1}
            \frac{\partial V_{\kappa}}{\partial \sigma_{\kappa,i}} \right)	- \kappa^T V_{\kappa}^{-1}
            \frac{\partial V_{\kappa}}{\partial \sigma_{\kappa,i}} V_{\kappa}^{-1} \kappa

        adding to de hessian : 
        
        .. math::
            \frac{\partial^2 \chi^2_{CS}}{\partial \kappa_j \partial \kappa_i} =  2  V_{\kappa}^{-1}

        .. math::
            \frac{\partial^2 \chi^2_{CS}}{\partial  \sigma_{\kappa,j} \partial  \sigma_{\kappa,i}} =
            -Tr\left( V_{\kappa}^{-1}  \frac{\partial V_{\kappa}}{\partial  \sigma_{\kappa,j}} V_{\kappa}^{-1}
            \frac{\partial V_{\kappa}}{\partial  \sigma_{\kappa,i}}  \right) + 2 \kappa^T V^{-1}_{\kappa}
            \frac{\partial V_{\kappa}}{\partial \sigma_{\kappa,i}} V_{\kappa}^{-1}
            \frac{\partial V_{\kappa}}{\partial \sigma_{\kappa,j}} V_{\kappa}^{-1} \kappa
        
        Parameters
        ----------
        derivatives : bool
           If gradient and hessian are needed.
        """
        sigma_kappa = self.sigma_kappa 
        # print(f'sigma_kappa : {sigma_kappa}')
        # sig = np.polyval(sigma_kappa, self.lambda_c_red) 
        if derivatives:
            sig, derivative_kappa_covmatrix = self.color_scatter(self.sigma_kappa, jac=True)  ##
            # vandermonde_matrix = np.vander(self.lambda_c_red, len(sigma_kappa)).T
        else :  ## 
            sig = self.color_scatter(self.sigma_kappa)  ##
            
        kappa = self.pull_function.model.pars['kappa_color'].full
        kappa_covmatrix = sig**2
        inv_kappa_covmatrix = 1/kappa_covmatrix
        
        coomatrix_inv_kappa_covmatrix = coo_matrix(np.diag(inv_kappa_covmatrix))
        kappa_index_start = self.pull_function.model.pars['kappa_color'].indexof()[0]
        total_parameters_number = int(np.sum([self.hess.shape])/2)
        
        row = kappa_index_start + coomatrix_inv_kappa_covmatrix.row
        col = kappa_index_start + coomatrix_inv_kappa_covmatrix.col
        extended_coomatrix_inv_kappa_covmatrix = coo_matrix((coomatrix_inv_kappa_covmatrix.data, (row, col)),
                                                            shape=(total_parameters_number, total_parameters_number))

        vc_cs = (inv_kappa_covmatrix.dot(kappa**2)).sum()
        self.chi2 += vc_cs
        self.detail['CS_hold'] = vc_cs

        if self.sigma_kappa_fit:
            log_det_cs = np.log(kappa_covmatrix).sum()
            self.chi2 += log_det_cs
            self.detail['CS_log_det'] = log_det_cs

        if derivatives:
            extended_coomatrix_inv_kappa_covmatrix = extended_coomatrix_inv_kappa_covmatrix.tocsc()
            self.hess += 2 * extended_coomatrix_inv_kappa_covmatrix
            self.grad += 2 * extended_coomatrix_inv_kappa_covmatrix.dot(self.rho)
                
            if self.sigma_kappa_fit:
                # derivative_kappa_covmatrix = (vandermonde_matrix * sig * 2).T
                trace_derivative_log_det_kappa = 1 * inv_kappa_covmatrix.dot(derivative_kappa_covmatrix)
                kappa_derivative_covmatrix_kappa = -1 * derivative_kappa_covmatrix.T.dot(
                    (inv_kappa_covmatrix * kappa)**2
                )

                grad_kappa = trace_derivative_log_det_kappa + kappa_derivative_covmatrix_kappa.ravel()
                trace_product = inv_kappa_covmatrix * derivative_kappa_covmatrix.T
                double_trace_log_det_kappa = - trace_product.dot(trace_product.T)
                derivative_covmatrix = derivative_kappa_covmatrix.T * inv_kappa_covmatrix * kappa
                double_product_derivative_covmatrix_kappa = 2*(derivative_covmatrix * inv_kappa_covmatrix).dot(
                    derivative_covmatrix.T
                )
                hess_kappa = coo_matrix(double_trace_log_det_kappa + double_product_derivative_covmatrix_kappa)
                hess_coo = self.hess.tocoo()

                hessian_combinaison = scipy.sparse.bmat([[hess_coo, None], [None, hess_kappa]], format='csc')
                
                self.grad = np.hstack((self.grad, grad_kappa))
                self.hess = hessian_combinaison
                
    def calibration_gaussian_prior(self, eta_covmatrix, derivatives=False):
        r"""        
        Compute calibration propagation component of the total :math:`\chi^2`:                          

        .. math::          
            \chi^2_{calib} = \eta^T V_{\eta}^{-1} \eta

        If derivatives are needed:
        adding to de gradient : 
        
        .. math::
            \frac{d \chi^2_{calib}}{d\eta} = V_{\eta}^{-1} \eta

        adding to de hessian : 
        
        .. math::
           \frac{d^2 \chi^2_{calib}}{d\eta^2} =  V_{\eta}^{-1}

        Parameters
        ----------
        eta_covmatrix : float or array
           Calibration uncertainty matrix.
        derivatives : bool
           If gradient and hessian are needed.
        """
        if type(eta_covmatrix) == float:
            eta_covmatrix = np.diag(np.ones((len(self.pull_function.model.bands))) * eta_covmatrix)
        else:
            eta_covmatrix = eta_covmatrix

        coomatrix_eta_covmatrix = coo_matrix(eta_covmatrix)
        inv_coomatrix_eta_covmatrix = scipy.sparse.linalg.inv(coomatrix_eta_covmatrix.tocsc()).tocoo()
        try:
            eta_index_start = self.pull_function.model.pars['photo_calibration'].indexof()[0]
        except KeyError:
            eta_index_start = self.pull_function.model.pars['eta_calib'].indexof()[0]
        if eta_index_start == -1:
            eta_index_start = 0

        row = eta_index_start + inv_coomatrix_eta_covmatrix.row
        col = eta_index_start + inv_coomatrix_eta_covmatrix.col

        extended_inv_cscmatrix_eta_covmatrix = coo_matrix((inv_coomatrix_eta_covmatrix.data, (row, col)),
                                                          shape=(self.N_PAR, self.N_PAR)).tocsc()
        v_eta = (extended_inv_cscmatrix_eta_covmatrix.dot(self.rho**2)).sum()
        
        self.chi2 += v_eta
        self.detail['C_eta'] = v_eta

        if derivatives:
            self.hess += 2 * extended_inv_cscmatrix_eta_covmatrix
            self.grad += 2 * extended_inv_cscmatrix_eta_covmatrix @ self.rho

    def get_constrains_derivatives(self, derivatives=False):
        r"""
        Evaluate the constrains :math:`C(\beta)` contribution the the :math:`\chi^2` as a quadratic penalty:

        .. math::                                                                                       
            \chi^2_{cons} = \mu_{pen} C(\beta)^T C(\beta)
            
        If derivatives are needed:
        adding to de gradient : 
        
        .. math::
            \frac{d \chi^2_{cons}}{d\beta} = 2 \mu_{pen} \frac{\partial C(\beta)}{\partial \beta_i}^T C(\beta)
        
        adding to de hessian : 
        
        .. math::
            \frac{d^2 \chi^2_{cons}}{d\beta^2} = 2 \mu_{pen} \left(\frac{\partial^2
            C(\beta)}{\partial \beta_j \partial \beta_i}^T C(\beta) +
            \frac{\partial C(\beta)}{\partial \beta_i}^T \frac{\partial C(\beta)}{\partial \beta_j} \right)

        Parameters
        ----------
        derivatives : bool
           If gradient and hessian are needed.
        """
        con = self.constrains(self.beta, jac=False)
        
        self.chi2 += (con**2).sum()
        self.detail['cons_qua'] = (con**2).sum()
        if len(self.pull_function.model.pars['X1'].free) != 0.:
            if derivatives:
                p_free = self.pull_function.VarianceModel.pars.free
                vcons, jacobian_cons, hessian_cons = self.constrains.var_x1(self.beta,
                                                                            gamma=p_free,
                                                                            jac=True)
                self.hess += hessian_cons.tocsc()
                self.grad += jacobian_cons

            else:
                vcons = self.constrains.var_x1(self.beta, jac=False)
            self.chi2 += vcons**2
            self.detail['x1_cons'] = vcons**2
            
        if derivatives:
            v, bloc_hessian_cons = self.constrains(self.beta, jac=True)
            if len(self.gamma) != 0.:
                bloc_hessian_cons.resize((bloc_hessian_cons.shape[0], self.N_PAR))
            self.hess += (2 * bloc_hessian_cons.T.dot(bloc_hessian_cons)).tocsc()
            self.grad -= 2 * bloc_hessian_cons.T.dot(v)


class ModelPulls:
    r"""
    Give the pulls of the model.
    For data point i, evaluated the model and the error model : :math:`\sigma` :

    .. math::
        pull = \frac{data_i - model_i(\beta)}{\sigma_i}


    .. note: ModelResiduals() should be able to accept either a model or a variance model !
    .. note: shorter code if model() could have the same signature regardless of jac=True or not

    Attributes
    ----------
    training_dataset : nacl.dataset.TrainingDataset
        Data set of photometric and spectroscopic observations.
    idxlc : int 
        Index where light curve data finish and spectra data starts.
    data : array 
        Data point array.
    N : int
        Number of data point.
    model : nacl.salt.
        Model
    chi2_normalization : None or float
        Normalization between modelization of photometric and spectroscopic datat
        :math:`\chi^2_{photo}` and :math:`\chi^2_{spectro}`.
    VarianceModel : None or nacl.models.variancemodels
        Variance model, error snake integration.      
    pars : FitParameters
        Model parameters.
    """
    
    def __init__(self, model, variance_model=None, chi2_normalization=None, fit_spec=True):
        r"""
        Model pull initialisation.

        Parameters
        ----------
        model : nacl.models.salt
            Model.
        chi2_normalization : None or float
            Normalization between modelization of photometric and spectroscopic datat
            :math:`\chi^2_{photo}` and :math:`\chi^2_{spectro}`.
        variance_model : None or nacl.models.variancemodels
            Variance model, error snake integration.      
        fit_spec : bool
            If spectroscopic data are fitted.
        """
        self.training_dataset = model.training_dataset
        self.idxlc = len(self.training_dataset.lc_data) 
        if fit_spec:
            self.data = np.hstack((self.training_dataset.lc_data, self.training_dataset.spec_data))
        else:
            self.data = self.training_dataset.lc_data
        self.N = len(self.data)
        self.model = model

        self.chi2_normalization = chi2_normalization
        if variance_model is not None:
            self.VarianceModel = variance_model
        else:
            self.pars = model.pars

    def __call__(self, beta, gamma=[], jac=True):
        r"""
        Evaluation of the pull.
        
        Parameters
        ----------
        beta : array
           Model parameters.
        gamma : array
           Error snake parameters.
        jac : bool
           If derivatives are required.

        Returns
        -------
        w_res : array
            Residuals multiply by the square root of the weight matrix.
        v : array
            Model evaluation.

        if jac is True :
        w_res : array
            Residuals multiply by the square root of the weight matrix.
        w_jac : scipy.sparse.coo_matrix
            Residuals derivatives multiply by the square root of the weight matrix.
        v : array
            Model evaluation.
        jacobian : scipy.sparse.coo_matrix
             Model derivative evaluation.
        variance : array or None
             Error snake evaluation.
        derivative_variance : scipy.sparse.coo_matrix or None
            Error snake derivative evaluation.
        """
        if jac:
            v, jacobian = self.model(beta, jac=True)
        else:
            v = self.model(beta, jac=False)
            
        res = self.data['Flux'] - v

        if len(gamma) != 0:
            variance = self.VarianceModel(gamma, beta)
        else:
            variance = self.data['FluxErr']**2

        ww = 1/np.sqrt(variance)
            
        if self.chi2_normalization is None:
            chi2_normalization = len(self.training_dataset.lc_data)/len(self.training_dataset.spec_data)
        else:
            chi2_normalization = self.chi2_normalization
            
        #print(f'chi2_normalization = {chi2_normalization}')
        ww[self.idxlc:] *= np.sqrt(chi2_normalization)        

        if len(variance.shape) == 1:
            w = scipy.sparse.dia_matrix((ww, [0]), shape=(self.N, self.N))
        else:
            w = scipy.sparse.csc_matrix(ww)

        w_res = w @ res
        if jac:
            w_jac = w @ jacobian  # -w @ jacobian #moins
            
            if len(gamma) != 0:
                variance, derivative_variance = self.VarianceModel(gamma, beta, jacobian=jacobian.tocoo())
                return w_res, w_jac, v, jacobian, variance, derivative_variance
            else:
                return w_res, w_jac, v, jacobian, variance, None
        
        if np.isnan(w@res).sum() > 0.:
            raise ValueError('There are np.nan in w@res')
        if np.isinf(w@res).sum() > 0.:
            raise ValueError('There are np.inf in w@res')
        return w_res, v



""" Simple tools to read descriptions of instruments.

This module provides a python equivalent to the SNFIT "Instrument"
class.
"""

from ..util.io import NTuple
import os.path as op
import sys

from . import constants, atmosphericmodel, spectrum
from .interpolation import Func1D
from .indextools import first_two_cols
from ..util import saltpath
from warnings import warn
from glob import glob


import numpy as np

#import pdb
import re
from functools import reduce

#######################################
# Flux and units
#######################################
OVERHC = 1. / (constants.C_LIGHT * constants.H * 1e17)


#######################################
# Reader for config and data files
#######################################
def read_transmission_files(filename):
    """Read the transmission files following the snfit_data convention.

    Handle correctly the headers in snfit_data in particular WAVELENGTH_IN_NM trigger conversion to angstrom
    """
    with open(filename) as fid:
        lines = fid.readlines()
    keys = {}
    comments = []
    i = 0
    for line in lines:
        if line[0] not in ['@', '#']:
            break
        if line[0] == '@':
            k = line[1:].split()
            keys[k[0]] = ' '.join(k[1:])
        else:
            comments.append(line)
        i = i+1
    vals = np.genfromtxt(lines[i:])


    if 'WAVELENGTH_IN_NM' in keys:
        vals[:, 0] *= 10
    return vals

def read_filter_wheel_file(filename):
    """ Read the filter wheel description
    """
    fid = open(filename)
    lines = fid.readlines()
    fid.close()
    fwheel = {}
    for line in lines:
        if line[0] == '#':
            continue
        values = line.split()
        if len(values) == 2:
            fwheel[values[0]] = fwheel.get(values[0], []) + [values[1]]
        elif len(values) == 3:
            fwheel[values[0]] = fwheel.get(values[0], []) + [values[2]]
        else:
            print(("Warning: Strange line in file %s" % filename))
            print(line)
    return fwheel


class Transmission(Func1D):
    """
    Special Func1D to handle filters
    
    Needed for them to know which Instrument they are from, and which Band they are
    """

    def __init__(self, *func_like, **kwargs):
        """
        By default the InstrumentName and Band are None
        InstrumentName: option to set the filter InstrumentName
        Band: option to set the filter Band
        """

        InstrumentName = kwargs.pop("InstrumentName", None)
        Band = kwargs.pop("Band", None)

        Func1D.__init__(self, *func_like, **kwargs)
        self.InstrumentName = InstrumentName
        self.Band = Band

    def __mul__(self, func2):
        """ Return a Func1D that will evaluate to the product of self
        and f2.
        """
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max

        InstrumentName = self.InstrumentName
        Band = self.Band
        
        if isinstance(func2, Func1D):
            x_min = max(self.x_min, func2.x_min)
            if func2.x_max != 0:
                x_max = min(self.x_max, func2.x_max)

        if  isinstance(func2, Transmission):
            x_min = max(self.x_min, func2.x_min)
            if func2.x_max != 0:
                x_max = min(self.x_max, func2.x_max)
            if func2.InstrumentName is not None:
                if InstrumentName is None:
                    InstrumentName = func2.InstrumentName
                else:
                    if func2.InstrumentName != InstrumentName:
                        raise ValueError("Multiplying transmissions from different instruments (%s, %s)" % (InstrumentName, func2.InstrumentName))

            if func2.Band is not None:
                if Band is None:
                    Band = func2.Band
                else:
                    if func2.Band != Band:
                        raise ValueError("Multiplying transmissions from different bands (%s, %s)" % (Band, func2.Band))
            
        return Transmission( lambda x: self.func(x) * Func1D( func2 ).func(x), x_min=x_min, x_max=x_max,
                             InstrumentName=InstrumentName,
                             Band=Band)

    #- Transmission multiplication is commutative and yields a Transmission
    __rmul__ = __mul__

    def __add__(self, func2):
        """ Return a Func1D that will evaluate to the sum of self
        and func2.
        """        
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max
        InstrumentName = self.InstrumentName
        Band = self.Band
        
        if isinstance(func2, Func1D):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )

        if  isinstance(func2, Transmission):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )
            if func2.InstrumentName is not None:
                if InstrumentName is None:
                    InstrumentName = func2.InstrumentName
                else:
                    if func2.InstrumentName != InstrumentName:
                        raise ValueError("Multiplying transmissions from different instruments (%s, %s)" % (InstrumentName, func2.InstrumentName))

            if func2.Band is not None:
                if Band is None:
                    Band = func2.Band
                else:
                    if func2.Band != Band:
                        raise ValueError("Multiplying transmissions from different bands (%s, %s)" % (Band, func2.Band))

        return Transmission( lambda x: self.func(x) + Func1D( func2 ).func(x), x_min=x_min, x_max=x_max,
                             InstrumentName=InstrumentName,
                             Band=Band)

    #- Transmission addition is commutative and yields a transmission
    __radd__ = __add__

    def __sub__(self, func2):
        """ Return a Func1D that will evaluate to the difference of self
        and func2.
        """        
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max
        InstrumentName = self.InstrumentName
        Band = self.Band
        
        if isinstance(func2, Func1D):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )


            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )
            if func2.InstrumentName is not None:
                if InstrumentName is None:
                    InstrumentName = func2.InstrumentName
                else:
                    if func2.InstrumentName != InstrumentName:
                        raise ValueError("Multiplying transmissions from different instruments (%s, %s)" % (InstrumentName, func2.InstrumentName))

            if func2.Band is not None:
                if Band is None:
                    Band = func2.Band
                else:
                    if func2.Band != Band:
                        raise ValueError("Multiplying transmissions from different bands (%s, %s)" % (Band, func2.Band))

        return Transmission( lambda x: self.func(x) - Func1D( func2 ).func(x), x_min=x_min, x_max=x_max,
                             InstrumentName=InstrumentName,
                             Band=Band)

    def __rsub__(self, func2):
        """ Return a Func1D that will evaluate to the difference of func2
        and self.
        """        
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max
        InstrumentName = self.InstrumentName
        Band = self.Band
        
        if isinstance(func2, Func1D):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )

        if  isinstance(func2, Transmission):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )
            if func2.InstrumentName is not None:
                if InstrumentName is None:
                    InstrumentName = func2.InstrumentName
                else:
                    if func2.InstrumentName != InstrumentName:
                        raise ValueError("Multiplying transmissions from different instruments (%s, %s)" % (InstrumentName, func2.InstrumentName))

            if func2.Band is not None:
                if Band is None:
                    Band = func2.Band
                else:
                    if func2.Band != Band:
                        raise ValueError("Multiplying transmissions from different bands (%s, %s)" % (Band, func2.Band))

        return Transmission( lambda x: Func1D( func2 ).func(x) - self.func(x), x_min=x_min, x_max=x_max,
                             InstrumentName=InstrumentName,
                             Band=Band)

    def __pow__(self, func2):
        """ Return a Func1D that will evaluates self ** func2
        """        
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max
        InstrumentName = self.InstrumentName
        Band = self.Band
        
        if isinstance(func2, Func1D):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )

        if  isinstance(func2, Transmission):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )
            if func2.InstrumentName is not None:
                if InstrumentName is None:
                    InstrumentName = func2.InstrumentName
                else:
                    if func2.InstrumentName != InstrumentName:
                        raise ValueError("Multiplying transmissions from different instruments (%s, %s)" % (InstrumentName, func2.InstrumentName))

            if func2.Band is not None:
                if Band is None:
                    Band = func2.Band
                else:
                    if func2.Band != Band:
                        raise ValueError("Multiplying transmissions from different bands (%s, %s)" % (Band, func2.Band))

        return Transmission( lambda x: self.func(x) ** (Func1D( func2 ).func(x)), x_min=x_min, x_max=x_max,
                             InstrumentName=InstrumentName,
                             Band=Band)

    def __rpow__(self, func2):
        """ Return a Func1D that will evaluates func2 ** self
        """        
        #- in this case, func2 would have default x_min and x_max 
        #- and 0 looses in the test just above
        x_min = self.x_min
        x_max = self.x_max
        InstrumentName = self.InstrumentName
        Band = self.Band
        
        if isinstance(func2, Func1D):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )

        if  isinstance(func2, Transmission):
            x_min = max( self.x_min, func2.x_min )
            if func2.x_max != 0:
                x_max = min( self.x_max, func2.x_max )
            if func2.InstrumentName is not None:
                if InstrumentName is None:
                    InstrumentName = func2.InstrumentName
                else:
                    if func2.InstrumentName != InstrumentName:
                        raise ValueError("Multiplying transmissions from different instruments (%s, %s)" % (InstrumentName, func2.InstrumentName))

            if func2.Band is not None:
                if Band is None:
                    Band = func2.Band
                else:
                    if func2.Band != Band:
                        raise ValueError("Multiplying transmissions from different bands (%s, %s)" % (Band, func2.Band))

        return Transmission( lambda x: (Func1D( func2 ).func(x)) ** self.func(x), x_min=x_min, x_max=x_max,
                             InstrumentName=InstrumentName,
                             Band=Band)
    
        
        
class FilterWheel(object):
    """ Handle the Filter transmissions

    The transmissions are Transmission instances
    """
    def __init__(self, path, filename, InstrumentName=None):
        # Filenames is a dictonnary with one key per band of the filterwheel
        filenames = read_filter_wheel_file(op.join(path, filename))
        self.transmissions = {}
        self.range = {}
        for band in filenames:
            transmission = read_transmission_files(op.join(path, filenames[band][0]))
            self.transmissions[band] = Transmission(transmission[:, 0], 
                                              transmission[:, 1],
                                              Band=band,
                                              InstrumentName=InstrumentName)
            self.range[band] = (self.transmissions[band].func.x.min(), 
                                self.transmissions[band].func.x.max())

    def __call__(self, band, **keys):
        """ Return the instrument transmission in band b

        Parameters:
        -----------
        band: string
              name of the photometric band

        focal_plane_position: not used in this class (see subclasses
                              such as RadiallyVariableFilterWheel for
                              position dependent Filters)
        """
        return self.transmissions[band]

    def get_filter_bandpass(self, band):
        """ Give the wavelength range covered by a filter

        Parameters:
        -----------
        band: string
              name of the photometric band
        """
        return self.range[band]


class RadiallyVariableFilterWheel(FilterWheel):
    """ Handle radially variable filters such as MegaCam filters.

    """
    def __init__(self, path, filename, InstrumentName=None):
        filenames = read_filter_wheel_file(op.join(path, filename))
        self.transmissions = {}
        self.range = {}
        self.radii = {}
#        for band in 'ugriyz':
        for band in filenames:
            self.transmissions[band] = [NTuple.fromtxt(
                op.join(path, f)) for f in filenames[band]]
            self.radii[band] = [float(nt.keys['MEASUREMENT_RADIUS'].replace('cm', '')) for nt in self.transmissions[band]]
            self.transmissions[band] = [Transmission(*first_two_cols(t),
                                               Band=band,
                                               InstrumentName=InstrumentName) for t in self.transmissions[band]]
            self.range[band] = (self.transmissions[band][0].func.x.min(),
                                self.transmissions[band][0].func.x.max())
            
    def _get_rad(self, rad, band):
        radii = self.radii[band]
        i = np.digitize([rad], radii)[0]
        imax = len(radii)-1
        if i > imax:
            i = imax
        i = i - 1
        r1, r2 = radii[i], radii[i + 1]
        w = (rad - r1) / (r2 - r1)
        if w < 0 or w > 1:
            raise ValueError('Radius %f is outside the focal plane' % rad)
        return i, w
    
    def __call__(self, band, focal_plane_position=0.):
        """ Return the instrument transmission in band b

        Parameters:
        -----------
        band: string
              name of the photometric band

        focal_plane_position: radius in cm
        """
        i, w = self._get_rad(focal_plane_position, band)
        return (self.transmissions[band][i + 1] * w) + (self.transmissions[band][i] * (1 - w))

class VariableFilterWheel(FilterWheel):
    """ Handle variable filters such as MegaCam filters.
    """
    def __init__(self, path, filename, InstrumentName=None):
        filenames = read_filter_wheel_file(op.join(path, filename))
        self.transmissions = {}
        self.range = {}
        self.pos = {}
        self.x = {}
        self.y = {}
        self.grid = {}
        for band in filenames:
            transmissions = [NTuple.fromfile(
                f) for f in glob(op.join(path, filenames[band][0]))]
            x = np.array([float(nt.keys['MEASUREMENT_X']) for nt in transmissions])
            y = np.array([float(nt.keys['MEASUREMENT_Y']) for nt in transmissions])
            self.pos[band] = np.rec.fromarrays([x,y], names=['x', 'y'])
            s = np.argsort(self.pos[band], order=['x', 'y'])
            self.pos[band] = self.pos[band][s]
            self.transmissions[band] = np.array([Transmission(*first_two_cols(transmissions[i]),
                                                     Band=band,
                                                     InstrumentName=InstrumentName) for i in s])
            self.x[band] = np.unique(self.pos[band]['x'])
            self.y[band] = np.unique(self.pos[band]['y'])
            self.range[band] = (self.transmissions[band][0].func.x.min(),
                                self.transmissions[band][0].func.x.max())
            self.transmissions[band] = self.transmissions[band].reshape((len(self.x[band]), len(self.y[band])))
            
    def _2D_interp(self, pos, band):
        i = np.digitize(pos[0], self.x[band]) - 1
        j = np.digitize(pos[1], self.y[band]) - 1
        x = self.x[band]
        y = self.y[band]
        if i > len(x) or i < 0 or j > len(y) or j < 0:
            raise ValueError('Radius %f, %f is outside the focal plane' % (pos[0], pos[1]))
        a, b = x[i], x[i+1]
        c, d = y[j], y[j+1]
        w_x = (pos[0] - a) / (b - a)
        w_y = (pos[1] - c) / (d - c)
        return i, j, w_x, w_y
    
    def __call__(self, band, focal_plane_position=(0.,0.)):
        """ Return the instrument transmission in band b

        Parameters:
        -----------
        band: string
              name of the photometric band

        focal_plane_position: radius in cm
        """
        i, j, w_x, w_y = self._2D_interp(focal_plane_position, band)
        r = self.transmissions[band]
        return ((r[i, j] * ((1 - w_x) * (1 - w_y))) +
                (r[i+1, j] * ((1 - w_y) * w_x)) + 
                (r[i, j + 1] * ((1 - w_x) * w_y)) +
                (r[i+1, j+1] * (w_x * w_y)))



class MagSys(object):
    """
    Handles the MagSys description
    """
    
    #+ Rajouter le truc de Marc de instmodel ici
    def __init__(self, magsys_name):
        """
        Read the magsys description from an snfit_data directory

        The reference spectrum is loaded in a spectrum under
        self.reference_spectrum

        Note that self.ZP is not the ZeroPoint as implemented in snfit/src/magsys.cc
        but the magnitude of the reference star in the filter.

        In order to get the ZeroPoint needed to calculate the magnitude, it is necessary
        to invoke self.ZeroPoint(ObserverFilter) where ObserverFilter is a Transmission instance.
        
        It can NOT be a simple Func1D instance, for it needs to know its InstrumentName and Band


        Parameters:
        -----------
        magsys_name should be described in fitmodel.cards pointed
        at by $SALTPATH
        """
        fitmodel_card = saltpath.read_fitmodel_card()
        if magsys_name == 'AB':
            self.pure_ab = True
            return
        else:
            self.pure_ab = False
        if not (magsys_name in fitmodel_card):
            raise ValueError("Couldn't find MagSys %s in your SALTPATH (%s)" % (magsys_name, saltpath.SALTPATH))

        magsys_filename = saltpath.SALTPATH + "/" + fitmodel_card[magsys_name]
            
        self.ZP = {}

        fh = open(magsys_filename, "r")
        for line in fh.readlines():
            #- remove comments and empty lines
            line = (re.sub("#.*$", "", line)).strip()        
            if len(line) == 0:
                continue
            if line.startswith("@"):
                group = re.search("^@(.+?)\s+(.+)", line)
                group = group.groups()
                if group[0] == "SPECTRUM":
                    self.reference_spectrum_path = saltpath.SALTPATH + "/" + group[1]
                    self.reference_spectrum = spectrum.load_ascii(self.reference_spectrum_path)
            #- This assumes that the magsys file is not exotic:
            #- if it is not a comment, an @ or an empty line, it is an instrument ZP
            else:
                data = line.split()
                if not (data[0] in list(self.ZP.keys())):
                    self.ZP[data[0]] = {}
                self.ZP[data[0]][data[1]] = float(data[2])

    def ZeroPoint(self, ObserverFilter):
        """
        Same implementation as snfit/src/magsys.cc
        
        ObserverFilter needs to be a Transmission instance 
        """
        if self.pure_ab:
            wl = np.linspace(ObserverFilter.x_min, ObserverFilter.x_max)
            reference_spectrum = spectrum.Spectrum(wl, constants.flux_ab(wl))
            return 2.5 * np.log10(reference_spectrum.IntegFlux(ObserverFilter, 0.))
        else:
            flux = self.reference_spectrum.IntegFlux(ObserverFilter, 0.)
            return 2.5 * np.log10(flux) + self.ZP[ObserverFilter.InstrumentName][ObserverFilter.Band]

    def mag(self, ObserverFilter, spectrum_list, spectrum_loader=lambda x : x):
        """
        Calculates the magnitude of spectra using an ObserverFilter that has to be
        a Transmission instance.

        If spectrum_loader is provided, uses it to read the data. Otherwise, assumes that 
        spectrum_list is a list of spectrum.Spectrum instance

        ObserverFilter is expected to be a list of ObserverFilters

        Examples:
        ---------
        load the spectra (if you run in test, where PTF09dnl data lives)
        >>> sn_name = "PTF09dnl"
        >>> sn_path = os.path.join("./", sn_name)
        >>> nt_sn = nacl.util.NTuple.fromtxt(os.path.join(sn_path, "%s_spec_merged.list" %  (sn_name)))
        >>> l_spectrum = []
        >>> for filename in nt_sn["filename"]:
               tmp_spectrum = spectrum.load_ntuple(os.path.join(sn_path, filename))
               l_spectrum.append(tmp_spectrum)

        Load the instrument and magsys
        >>> instrument = instruments.InstrumentModel("SWOPE")
        >>> magsys = instruments.MagSys("VEGA")
        >>> observer_filter = instrument.EffectiveFilterByBand("g")


        Possible to use only one filter and spectrum
        >>> magsys.mag(observer_filter, l_spectrum[3])
        array([(-19.94143291159957,)],
              dtype=[('g', '<f8')])

        Possible to calculate the mags of the full spectrum list
        >>> magsys.mag(observer_filter, l_spectrum)
            array([(-19.829897425083587,), (-19.920241486850074,),
                   (-19.932614171822166,), (-19.94143291159957,),
                   (-19.88517517546827,), (-19.889984971628934,),
                   (-19.84623774750333,), (-19.835170085307567,),
                   (-19.69677647212223,), (-19.560002435004172,),
                   (-19.357319526502515,), (-18.99707016591549,),
                   (-18.905751902417677,), (-18.895062900516358,),
                   (-18.823659018820486,), (-18.82680933234854,),
                   (-18.59134345034653,), (-18.275082295370318,),
                   (-18.106126338432183,), (-18.0919526101773,),
                   (-17.823886134044365,), (-17.787802645726963,), (-17.516200221607,),
                   (-17.37836748599415,), (-17.38841134403154,),
                   (-17.363730161058008,), (-17.3197802772009,),
                   (-17.263928780299743,), (-17.207123524932392,),
                   (-17.19225272676723,)],
                  dtype=[('g', '<f8')])

        Possible to load the spectra inside of the mag calculation, with a spectrum_loader       
        >>> os.chdir(sn_path)
        >>> magsys.mag(observer_filter, nt_sn["filename"], spectrum_loader=spectrum.load_ntuple)
            array([(16.01232254417715,), (15.92860061849749,), (15.921179105837751,),
                   (15.913653784190643,), (15.977232537932833,), (15.971919469595042,),
                   (16.022310674641226,), (16.033430436507768,), (16.19766767411273,),
                   (16.343111225705233,), (16.534715100013916,), (16.948693007225014,),
                   (17.053167003173723,), (17.066288834748185,), (17.15258042408282,),
                   (17.14327408888971,), (17.414470723339335,), (17.764370290920283,),
                   (17.95434843068567,), (17.971541356898307,), (18.211643011663316,),
                   (18.21272014546581,), (18.52436971411903,), (18.650500460393452,),
                   (18.639698400983757,), (18.651509477256596,), (18.622885337729812,),
                   (18.746686051787446,), (18.82050490149692,), (18.763722405091343,)],
                  dtype=[('g', '<f8')])

        And also possible to loop over filters instances:
        
        """
        # Syntaxic sugar to allow the easy use of mag on only one spectrum or only one transmission
        if isinstance(spectrum_list, spectrum.Spectrum):
            spectrum_list = [spectrum_list]
        if isinstance(ObserverFilter, Transmission):
            ObserverFilter = [ObserverFilter]

            
        res = np.zeros(len(spectrum_list), dtype=[(of.Band, float) for of in ObserverFilter])
        for _sp, sp in enumerate(spectrum_list):
            s = spectrum_loader(sp)
            res[_sp] = tuple(s.IntegFlux(of, 0.) for of in ObserverFilter)
        for of in ObserverFilter:
            Zp = self.ZeroPoint(of)
            res[of.Band] = -2.5 * np.log10(res[of.Band]) + Zp
        return res
        

class InstrumentModel(object):
    """Handle instrument description.

    The way we handle transmissions needs some convention to be clearly stated out here.

    A transmission is a value that describes how a given quantity (say photons)
    is transfered, without changing the nature of the quantity (i.e. yielding photons in this case).
    It is a ratio between 0 and 1. 

    An efficiency is a value that describes how a given quantity is transformed
    into an other quantity. For example how a filter+detector transforms photons
    in electrons. Or in ADUs.

    Since photons are the correct way to deal with flux in cosmology analysis,
    we decide to only consider transmissions in photons per photons, and
    efficiencies in electrons per photons. An option will be made available for
    transforming photons in ADUs in order prevent the mistake inherent to the
    fact that a gain is in e/ADUs (one thus has to *divide* the efficiency in
    electrons per photons by the gain in order to get ADUs per photons)

    We expect the spectra to be in photons/A/cm^2, but they are usually provided
    in ergs/A/cm^2, which is why the Spectrum class implements a
    s.photon method that provides the Func1D giving the spectrum of the object
    in photons/A/cm^2
    

    The functions that provide those different values are:

    self.filters(band, focal_plane_position): the pure filter transmission in photons per photons

    get_transmission(band): provides the instrument transmission (between 0 and 1) in photons per photons 
                            includes mirror reflectivity and optics transmission

    get_efficiency(band): provides the efficiency of transformation between
                          photons and electrons includes the instrument
                          transmission, the quantum efficiency of the detector
                          and the area of the mirror

    get_efficiency(band, unit="ADU"): provides the efficiency of transformation between photons and electrons

    EffectiveFilterByBand(band): for similarity with snfit. Provides the
    "transmission" in ADUs/photons, or using the semantic defined above, the
    efficiency in ADUs per photons

    """
    def __init__(self, name=None, path=None):
        """ Read the instrument description from an snfit_data
        directory.

        Parameters:
        -----------
        name: the name of an instrument described in fitmodels.card in the
        $SALTPATH

        path: The path to an instrument directory. This is for convenience.
              if a name is provided, the filter will have this name, otherwise, 
              it will have the default name "default_name"

        """
        if path is not None:
            #- here we are in a special case, where the user knows what he is doing
            #- he is trying to access a special instrument file directly, for testing
            if name is None:
                #- We need a name for the InstrumentModel instance, no matter what
                name = "default_name"                
            instrument_path = path
        else:
            #- here no special path is provided. Name should be provided
            #- Follwoing the Instrument file system, we should have a 
            #- directory labelled by the instrument name
            if name is None:
                raise ValueError("Could not create instrument without name")            
            fitmodels = saltpath.read_fitmodel_card()            
            try:
                instrument_path = fitmodels[name]
            except KeyError:
                raise ValueError('Unknown instrument %s' % name)
            instrument_path = op.join(saltpath.SALTPATH, instrument_path)

        instrument_file = op.join(instrument_path, 'instrument.cards')

        self.name = name
        self.path = instrument_path
        self.cards = saltpath.read_card_file(instrument_file)

        self.mirror_reflectivity = self._load('MIRROR_REFLECTIVITY')
        self.optics_transmission = self._load('OPTICS_TRANS')
        self.ccd_qe = self._load('QE')
        
        self.area = float(self.cards['MIRROR_AREA'])  # in cm^2

        if 'GAIN' in self.cards:
            self.gain = float(self.cards['GAIN'])
        else:
            self.gain = 1.
        
# deprecated
#        conv = self.area
#        self.open_transmission = (self.mirror_reflectivity *
#                                  self.optics_transmission *
#                                  self.ccd_qe * conv)
        if "RADIALLY_VARIABLE_FILTERS" in self.cards:
            self.filters = RadiallyVariableFilterWheel(
                self.path, self.cards['RADIALLY_VARIABLE_FILTERS'], InstrumentName=self.name)
        elif "VARIABLE_FILTERS" in self.cards:
            self.filters = VariableFilterWheel(
                self.path, self.cards['VARIABLE_FILTERS'], InstrumentName=self.name)
        elif "FILTERS" in self.cards:
            self.filters = FilterWheel(self.path, self.cards['FILTERS'], InstrumentName=self.name)

        self.atm_trans = None
        self.atm_model = None
        if "ATMOSPHERIC_TRANS" in self.cards:
            self.atm_trans = self._load("ATMOSPHERIC_TRANS")
        elif "ATMOSPHERIC_MODEL" in self.cards:
            self.atm_model = getattr(atmosphericmodel, self.cards['ATMOSPHERIC_MODEL'])
            

    def _load(self, key):
        """ Load component of the instrument

        Parameters:
        -----------
        key: string
             The name of the component in the card file
        Return:
        -------
        A Transmission with the transmission of the component (or 1 if the
        component is not found in the card file)
        """
        trans = Transmission(1.)
        if key in self.cards:
            try:
                trans = Transmission(float(self.cards[key]))
            except ValueError:
                tr = []
                for path in self.cards[key].split():
                    data = read_transmission_files(op.join(self.path, path))
                    tr.append(Transmission(data[:, 0], data[:, 1]))
                trans = reduce(lambda x, y: x*y, tr)
        return trans

    def get_transmission(self, band, **keys):
        """ Return the instrument transmission in photons per photons
        as a Transmission (derived from Func1D) instance

        It should be in the interval [0., 1.]
        
        This accounts for:
            the filter transmission : accessible via self.filters(band, focal_plane_position)
            the mirror reflectivity : in self. mirror_reflectivity
            the optic transmission  : in self.optics_transmission

        Parameters:
        -----------
        band: string
              The name of the photometric band 

        focal_plane_position: parameter to be passed to filterwheel.
        """
# deprecated        
#        if band == 'open':
#            return self.open_transmission
#        else:
#            return (self.filters(band, focal_plane_position) *
#                    self.open_transmission)
        return self.filters(band, **keys) * self.mirror_reflectivity * self.optics_transmission

    def get_efficiency(self, band, unit="e-", **keys):
        """
        Returns the instrument efficiency in e- per photons.
        
        This accounts for 
            the instrument transmission: provided by get_transmission (includes filter transmission, mirror reflectivity and optics transmission)
            the quantum efficiency : in self.qe
            the mirror area : in self.area        
         
        if unit="e-", the default that's all.
        if unit="ADU", it also includes the gain
        """

        instrument_transmission = self.get_transmission(band, **keys)
        electron_efficiency = instrument_transmission * self.ccd_qe * self.area
        
        if unit == "e-":
            return electron_efficiency
        elif unit == "ADU":
            return electron_efficiency * (1./self.gain)
        else:
            print(("unit %s not implemented" % (unit)))
            raise NotImplementedError
 
    def get_efficiency_at_airmass(self, band, airmass=1.25, reference_airmass=1.25, **keys):
        """
        Calculates the instrument efficiency in e- per photons, in cluding the atmospheric transmission

        This needs the instrument description to provide an atmospheric effective transmission curve.
        """
        
        efficiency = self.get_efficiency(band, **keys)
        if self.atm_trans is not None:
            return efficiency * self.atm_trans ** (airmass / reference_airmass)
        elif self.atm_model is not None:
            return efficiency * self.atm_model(airmass=airmass)
        else:
            warn(Warning("Instrument has no separate atmospheric transmission curve"))
            return efficiency
        
    def EffectiveFilterByBand(self, band, **keys):
        """
        For easy comparison with snfit: returns the effective transmission (i.e. efficiency) in ADUs 
        of a given filter.

        This is snfit idiosyncratic, and therefore also includes airmass at
        airmass 1.25 when available, as well as the gain.
        """        

        return  self.get_efficiency_at_airmass(band, unit="ADU", **keys)



if __name__ == "__main__":
    #- Implements few tests that I feel better if I can run from time to time
    '''
    x = Transmission(np.arange(1., 10., 0.1), np.arange(1., 10., 0.1))
    y = Transmission(np.arange(1., 10., 0.1), np.arange(1., 10., 0.1)+10.)    
    
    print "x(4.) + 1.    = %f     (should be 5.)" % ((x + 1.)(4.))
    print "1.    + x(4.) = %f     (should be 5.)" % ((1. + x)(4.))

    print "x(4.) - 1.    = %f     (should be 3.)" % ((x - 1.)(4.))
    print "1.    - x(4.) = %f     (shoudld be -3.)" % ((1. - x)(4.))


    print "x(4.) * 2.    = %f     (should be 8.)" % ((x * 2.)(4.))
    print "2.    * x(4.) = %f     (should be 8.)" % ((2. * x)(4.))
    
    print "x(4.) ** 2.    = %f     (should be 16.)" % ((x ** 2.)(4.))
    print "2.    ** x(4.) = %f     (should be 16.)" % ((2. ** x)(4.))

    print "(x + y)(2.)    = %f     (should be 14.)" % ((x + y)(2.))
    print "(y + x)(2.)    = %f     (should be 14.)" % ((y + x)(2.))
    print "(x * y)(2.)    = %f     (should be 24.)" % ((x * y)(2.))
    print "(y * x)(2.)    = %f     (should be 24.)" % ((y * x)(2.))
    print "(x ** y)(2.)   = %f     (should be 4096.)" % ((x ** y)(2.))
    print "(y ** x)(2.)   = %f     (should be 144.)" % ((y ** x)(2.))    
    '''
    instrument = InstrumentModel('MEGACAM6')
    
